-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Diagrams.Rules
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- data types for reduction and transformation rules, and 
-- several function and instances for the type
--
--
-----------------------------------------------------------------------------

module LRSX.Diagrams.Rules where
import Data.List
import Data.Maybe
--
import Debug.Trace
--
import LRSX.Language.Syntax
import LRSX.Unification.UnificationProblem
import LRSX.Unification.Unification
import LRSX.Language.Constraints

-- Reduction / Transformation Rules
data Rule a = a :==>: a
 deriving(Eq,Ord)

 
-- some helper functions to get the rule of an Either String String-rulename 
getRule (Left rule) = rule
getRule (Right rule) = rule
getSide (Left  _) (l,r)  = l
getSide (Right _) (l,r) = r
getOtherSide (Left  _) (l,r)  = r
getOtherSide (Right _) (l,r) = l
-- lookup a rule from a list of rules by its rulename
lookupRule key list = lookup key [(ruleName r,r) | r <- list] 
 
 
-- reverse a rule by swapping its lhs and rhs
reverseRule :: Rule a -> Rule a
reverseRule (a :==>: b) = (b :==>: a)
 
instance Show a => Show (Rule a) where
 show (a :==>: b) = show a ++ " ==> " ++ show b
 
instance HasFnSymbol a => HasFnSymbol (Rule a) where
 getFnSymbols (a :==>: b) = nubFnSyms $ getFnSymbols a ++ getFnSymbols b 
 
instance ApplySolution a => ApplySolution (Rule a) where
 applySolution s (a :==>: b) = (applySolution s a) :==>: (applySolution s b) 
 -- applySolutionWeaken s (a :==>: b) = (applySolutionWeaken s a) :==>: (applySolutionWeaken s b) 

instance HasMetaVars a => HasMetaVars (Rule a) where
 computeMetaVars (a :==>: b) = addMetaVars b (computeMetaVars a)
 
instance HasMetaContexts a => HasMetaContexts (Rule a) where
 computeMetaContexts (a :==>: b) = (computeMetaContexts a) ++ (computeMetaContexts b)


type ExprRule = Rule Expression
type RuleName = (Bool,String,Int,Bool)  -- (StandardReduction,Name,Variantnumber,TransitiveClosure?)
showRuleName (b,str,i,tc)
 | b = "SR,"++ str ++ (show i) ++ (if tc then ",+" else "")
 | otherwise = str ++ (show i) ++ (if tc then ",+" else "")

-- the data type for a rule with constraints:
data ConstrainedRule  = 
    ConstrainedRule {
        rule              :: ExprRule         -- ^ a rule where left and right hand side are expressions
       ,nonEmptyCtxts     :: [Expression]            -- ^ a list of contexts, which are non-empty
       ,nonEmptyEnvs      :: [EnvVar]                -- ^ a list of Environment-Variables which are non-empty
       ,nonCapCon         :: NonCaptureConstraints   -- ^ non-capture constraints
       ,priority          :: Int                     -- ^ a priority of the rules (for later use while closing diagrams)
       ,answerRule        :: Bool                    -- ^ a flag whether the rule is an answer rule
       ,standardReduction :: Bool                    -- ^ a flag whether the rule is a standard reduction of the program calculus
       ,ruleName          :: RuleName                -- ^ the name of the rule as a tuple 
    }
  deriving(Show)

identity = 
    ConstrainedRule {
        rule              = (MetaExpression (emptySubstitution) "@@ID") :==>:  (MetaExpression (emptySubstitution) "@@ID")
       ,nonEmptyCtxts     = []
       ,nonEmptyEnvs      = []
       ,nonCapCon         = []
       ,priority          = 0
       ,answerRule        = False
       ,standardReduction = False
       ,ruleName          = (False,"Identity",1,False)
    }

  
  
standardReductions :: [ConstrainedRule] -> [ConstrainedRule]
standardReductions = filter  standardReduction
answerReductions :: [ConstrainedRule] -> [ConstrainedRule]
answerReductions = filter  answerRule
  
-- reverse a constrainted rule by swapping its lhs and rhs
reverseConstrainedRule :: ConstrainedRule -> ConstrainedRule  
reverseConstrainedRule crule = crule{rule= reverseRule (rule crule)}


isAnswerRuleCommand answerRules (Single leftrule rightrule _) =
 case (leftrule) of
   Left (sr,name,num,p) -> case 
                            lookupRule (True,name,num,p) answerRules of
                             Nothing -> case lookupRule (False,name,num,p) answerRules of 
                                           Nothing -> False
                                           Just _ -> True
                             Just _ -> True
   _ -> False                             
   
   
-- overlap commands
data Command = 
        Single 
         (Either RuleName RuleName) 
         (Either RuleName RuleName) 
         IgnoredRulesForClosing                      
                                                             
      | All 
         (Either RuleName RuleName) 
         IgnoredRulesForOverlap
         IgnoredRulesForClosing
         
      | UNION 
         RuleName
         SetOfRules
-- Single (Dir rulename1) (Dir rulename2) means that 
-- that rules named rulename1 and rulename2 are overlapped
-- where Dir is Left or Right and determines the lhs or rhs    
-- All (Dir rulename) rulenames means that the lhs or rhs of the                            
-- the rule named rulenamed has to be overlapped with the lhs's
-- of all standard reduction rules except for the rules in rulenames
-- the lists IgnoredRulesForClosing / IgnoredRulesForOverlap have the following
-- meaning:
 deriving(Show)
type IgnoredRulesForClosing = [RuleName]  -- Names of such rules which should not be used during the search for closing diagrams
type IgnoredRulesForOverlap = [RuleName]  -- Names of such rules which should not be overlapped, when using the All-commands
type SetOfRules = [RuleName]                                                              
-- unfoldAll eliminates the  All commands from a list of commands by replacing it 
-- with the corresponding Single-Commands
unfoldAll :: [ConstrainedRule] -> [ConstrainedRule] -> [Command] -> [Command]
unfoldAll stdRules answerRules ((UNION a b):xs) =(UNION a b):(unfoldAll stdRules answerRules xs)
unfoldAll stdRules answerRules ((Single x y ignoreForClosing):xs) =(Single x y ignoreForClosing):(unfoldAll stdRules answerRules xs)
unfoldAll stdRules answerRules ((All x ys ignoreForClosing):xs)   
        = [(Single x r ignoreForClosing) | r' <- stdRules, (ruleName r') `notElem` ys, let r = Left (ruleName r') ] 
        ++ [(Single x r ignoreForClosing) | r' <- answerRules, let r = Left (ruleName r')] ++ (unfoldAll stdRules answerRules xs)
        
        
unfoldAll stdRules _ [] = []

-- ============================================================================================================
-- Interface to compute critical pairs for rule overlaps, given as overlap commands
-- 
--

data OverlapProblem = 
     OverlapProblem {
          rules         :: [ConstrainedRule]    -- ^ all given rules
         ,command       :: Command              -- ^ the overlap command
         ,freshNames    :: [String]             -- ^ fresh names
         ,funInfo       :: FunInfo              -- ^ FunInfo for the function symbols occuring in the problem
         ,criticalPairs :: [CriticalPair]       -- ^ the output of overlapping will fill the list of critical pairs
         ,isCommuting   :: Bool                 -- ^ determines whether the overlap is a commuting or a forking case
         ,opLogging     :: String               -- ^ used for logging
         ,answerOverlap :: Bool                 -- ^ overlapping for answer diagram
     }
     
data CriticalPair =
     CriticalPair {
         middleExpression :: Expression,        -- ^ the unified expression 
         leftExpression :: Expression,          -- ^ the left expression of the critical pair
         rightExpression :: Expression,         -- ^ the right expression of the critical pair
         ruleNameMiddleToLeft :: Either RuleName RuleName,  --  ^ the rule from middle to left given as Dir rulename, where Dir is Left/Right
         ruleNameMiddleToRight :: Either RuleName RuleName, --  ^ the rule from middle to right given as Dir rulename, where Dir is Left/Right
         unifier :: Problem, --  ^ the unifier, it is necessary to keep it, since it contains the Delta-Constraints
                            --    which are necessary for closing overlaps, i.e. the NCCs in Delta3 and the constraint equations in Delta4
         leftRuleOrig :: ConstrainedRule    , -- ^ the original but renamed and oriented rule 
         rightRuleOrig :: ConstrainedRule,    -- ^ the original but renamed and oriented rule 
         cpLogging     :: String -- ^ used for logging
         
     }

-- execute a sequence of overlap problems
-- the output is a again an overlap problem
-- where the critical pairs are filled

executeCommands :: [OverlapProblem] -> [OverlapProblem]
executeCommands overlapProblems = concatMap executeCommand overlapProblems


--
-- execute a single overlap command:
--


executeCommand :: OverlapProblem -> [OverlapProblem]
--
-- translate All-commands into a sequence of Single-commands 
executeCommand pb@(OverlapProblem{command = All rule exrules exrulesClosing})  =
    executeCommands [pb{command = cmd} | cmd <- (unfoldAll (standardReductions (rules pb)) (answerReductions (rules pb)) [All rule exrules exrulesClosing])]
--
-- Single-command, first lookup the corresponding rules:
executeCommand pb@(OverlapProblem{command = Single rule1 rule2 ignoreForClosing, freshNames = fresh, funInfo = fnSyms})
 | Just crule1  <- lookupRule (getRule rule1) (rules pb)
 , Just crule2  <- lookupRule (getRule rule2) (rules pb) = 
  let       
        -- make both rules fresh
        ([freshcrule1,freshcrule2],fresh') = makeAllFresh [crule1,crule2] fresh
        (l  :==>: r)         = rule freshcrule1
        (l' :==>: r')        = rule freshcrule2
        leftruleRenamedAndOriented = case rule1 of
                                      Left _ -> freshcrule1
                                      Right _ -> reverseConstrainedRule freshcrule1
        rightruleRenamedAndOriented = case rule2 of
                                       Left _ -> freshcrule2
                                       Right _ -> reverseConstrainedRule freshcrule2
        -- generate expression equations
        equations            = Gamma { varLiftEq = []
                                     , expressionEq = [getSide rule1 (l,r) :=?=: getSide rule2 (l',r')]
                                     , environmentEq = []
                                     , bindingEq = []
                               } 
        -- generate constraints:                               
        contextConstraints     = nonEmptyCtxts freshcrule1 ++ nonEmptyCtxts freshcrule2
        environmentConstraints = nonEmptyEnvs freshcrule1 ++ nonEmptyEnvs freshcrule2
        captureConstraints     = nonCapCon freshcrule1 ++ nonCapCon freshcrule2 ++ computeConstraints l ++ computeConstraints r ++ computeConstraints l' ++ computeConstraints r'
        -- built the unification problem
        problem                = Problem {
                                     solution       = []
                                    ,gamma          = equations
                                    ,delta          = Delta {delta1=contextConstraints,delta2=environmentConstraints,delta3=captureConstraints,delta4=emptyGamma}
                                    ,fresh          = fresh'
                                    ,functions      = fnSyms
                                    ,inputGamma     = equations
                                    ,inputDelta     = Delta {delta1=contextConstraints,delta2=environmentConstraints,delta3=captureConstraints,delta4=emptyGamma}
                                    ,constantDelta  = emptyDelta
                                    ,alphaBase      = []
                                    ,mpLogging = []
                                    ,expandAllContexts = False -- should be parametrized
                                }
        -- unify the problem, and single out only successful unifiers
        unificationResultTree =  unification problem
        unificationResult = solvedAndOpenLeaves  unificationResultTree
        -- components of the critical pairs:
        left                 = getOtherSide rule1 (l,r)
        right                = getOtherSide rule2 (l',r')
        middle               = getSide rule1 (l,r)        
        computedCriticalPairs = map (computeCriticalPair middle left right rule1 rule2 leftruleRenamedAndOriented rightruleRenamedAndOriented) unificationResult
        rulesAfterRemoveIgnored = [r | r <- (rules pb), (ruleName r) `notElem` ignoreForClosing]
        isCommutingOrForking = case (rule1,rule2) of
                                 (Left _, Left _) -> False -- Forking
                                 (Right _, Left _) -> True -- Commuting
   in -- check whether the problem fulfills the weak DVC
      case check problem of 
        Nothing ->  [pb{rules = rulesAfterRemoveIgnored, criticalPairs = computedCriticalPairs,isCommuting=isCommutingOrForking}]
        Just err -> error err
 
-- Otherwise, if rules are not found, stop with an error    
executeCommand pb@(OverlapProblem{command = Single rule1 rule2 ignoreForClosing, freshNames = fresh, funInfo = fnSyms})
 | otherwise = error $ "unknown rules" ++ show rule1 ++ " " ++ show rule2 
    
    
-- a simple helper function to generate the critical pair    
computeCriticalPair middle left right ruleMtoL ruleMtoR leftOrig rightOrig (Node p _) = error $ "PANIC: found open leaf " ++ show p       
computeCriticalPair middle left right ruleMtoL ruleMtoR leftOrig rightOrig (Solved pb) = 
 computeCriticalPairFromProblem middle left right ruleMtoL ruleMtoR leftOrig rightOrig pb

computeCriticalPairFromProblem middle left right ruleMtoL ruleMtoR leftOrig rightOrig pb =     
     CriticalPair {
          middleExpression     = applySolution (solution pb) middle
         ,leftExpression        = applySolution (solution pb) left
         ,rightExpression        = applySolution (solution pb) right
         ,ruleNameMiddleToLeft  = ruleMtoL
         ,ruleNameMiddleToRight = ruleMtoR
         ,unifier = pb
         ,leftRuleOrig = leftOrig 
         ,rightRuleOrig = rightOrig
         ,cpLogging = let showEither (Left str)  = ".<=" ++ showRuleName str ++ "="
                          showEither (Right str) = "=" ++ showRuleName str ++ "=>."
                      in
                       unlines ["Overlapping:"  ++ showEither ruleMtoL ++ " with " ++ showEither ruleMtoR
                               , "Rule: " ++  show leftOrig
                               , "with"
                               , "Rule: " ++ show rightOrig
                               , "yields the critical pair triple:"
                               , "left:\t"   ++ (show $ applySolution (solution pb) left)
                               , "middle:\t" ++ (show $ applySolution (solution pb) middle)
                               , "right:\t"  ++ (show $ applySolution (solution pb) right)
                               , "and delta:" ++ (show (constantDelta pb))
                              ]
     }
     
expandAllContextsCriticalPair  cp =
    let oldPB = unifier cp
        unificationResultTree = unification oldPB{expandAllContexts=True}
        unificationResult = solvedAndOpenLeaves  unificationResultTree
        -- components of the critical pairs:
        computedCriticalPairs = map (computeCriticalPair
                                        (middleExpression cp) 
                                        (leftExpression cp) 
                                        (rightExpression cp) 
                                        (ruleNameMiddleToLeft cp) 
                                        (ruleNameMiddleToRight cp) 
                                        (leftRuleOrig cp) 
                                        (rightRuleOrig cp))
                                    unificationResult 
                              
    in computedCriticalPairs

    
expandAllContextsOverlapProblem op =
    let [cp] = criticalPairs op
        newCps = expandAllContextsCriticalPair cp
        flattened  = [op{criticalPairs = [cp]}  | cp <- newCps]      
    in flattened        


     
-- 
-- given a list of constrained rules and a list of fresh names,
-- compute renamed rules, and the remaining fresh names
makeAllFresh :: [ConstrainedRule] -> [String] -> ([ConstrainedRule],[String])
makeAllFresh [] fresh = ([],fresh)
makeAllFresh (r:rs) fresh =
 let (r',fresh') = makeFresh r fresh
     (rs',fresh'') = makeAllFresh rs fresh'
 in (r':rs',fresh'')
 
makeFresh :: ConstrainedRule -> [String] -> (ConstrainedRule,[String]) 
-- DEBUG  
makeFresh crule fresh 
  | (l :==>: _) <- rule crule
  , Just err <- checkWeakDVC l   = error $ "makeFresh: WeakDVC does not hold (1):" ++ show l ++ err
  | (r :==>: _) <- rule crule
  , Just err <- checkWeakDVC r   = error $ "makeFresh: WeakDVC does not hold (2):" ++ show r ++ err  
-- END DEBUG  
makeFresh ConstrainedRule{ruleName=name,rule=theRule,nonEmptyCtxts=ctxts,nonEmptyEnvs=envs,nonCapCon=capcon,priority=prio,standardReduction=st,answerRule=ar} fresh =
 let metavars          = (computeMetaVars theRule)
                            `combineMetaVars` (computeMetaVars (Delta {delta1=ctxts,delta2 =envs, delta3 =capcon,delta4=emptyGamma}))
     alllamlet         = nub (liftvars metavars)
     (mLamLetVars1,f2) = zipWithFresh (\x y -> (MapVarLift $ MetaVarLift emptySubstitution x :|-> MetaVarLift emptySubstitution ("X" ++ y))) alllamlet fresh
     (mEnvVars,f3)     = zipWithFresh (\x y -> MapEnv $  x :|-> metaEnv (Env emptySubstitution ("E" ++ y))) (map (Env emptySubstitution) $ nub $ envvars metavars) f2
     mEnvVarsShrink    = zip (map (Env emptySubstitution) $ nub $ envvars metavars) (map (\x -> (Env emptySubstitution ("E" ++ x))) f2)
     (mExprVars,f4)    = zipWithFresh (\x y -> MapExpr $ MetaExpression emptySubstitution x :|-> MetaExpression emptySubstitution ("S" ++ y)) (nub $ exprvars metavars) f3
     (mCtxtVars,f5)    = zipWithFresh (\(MetaCtxt sub srt x) y -> MapExpr $ MetaCtxt sub srt x :|-> MetaCtxt sub srt ((repr srt) ++y)) (nub $ computeMetaContexts theRule) f4
     (mChainVars,f6)   = zipWithFresh (\x y -> (x,y)) (nub $ chainvars metavars) f5
     substitution = (mLamLetVars1 ++ mExprVars ++ mEnvVars ++ mCtxtVars)
     substituteEnvs subs []      = []
     substituteEnvs subs (z:zs)
        | Just y <- lookup z subs = y:(substituteEnvs subs zs)
        | otherwise               = z:(substituteEnvs subs zs)
 in 
    ((ConstrainedRule {ruleName=name
                      ,priority=prio
                      ,standardReduction=st
                      ,answerRule=ar
                      ,rule = subsEE mChainVars (applySolution substitution theRule)
                      ,nonEmptyCtxts = map (subsEE mChainVars) $ applySolution substitution ctxts
                      ,nonEmptyEnvs = substituteEnvs mEnvVarsShrink envs
                      ,nonCapCon = map (subsEE mChainVars) $ applySolution substitution capcon
                      }),f6)
                    
-- a type class for substitution of EE-variables, which is mainly used
-- during fresh renaming of rules
class SubsEE a where
 subsEE :: [(String,String)] -> a -> a
instance (SubsEE a, SubsEE b) => SubsEE (a, b) where
 subsEE subs (a,b) = (subsEE subs a,subsEE subs b)
instance (SubsEE a, SubsEE b) => SubsEE (Either a b) where
 subsEE subs (Left a) = Left (subsEE subs a)
 subsEE subs (Right a) = Right (subsEE subs a)
 
instance SubsEE a => SubsEE (Rule a) where
 subsEE subs (a :==>: b) = (subsEE subs a) :==>: (subsEE subs b)
instance SubsEE ArgPos where
 subsEE subs (Binder v e) = Binder v (subsEE subs e)
 subsEE subs (VarPos v ) = VarPos v
instance SubsEE Expression where
 subsEE subs (VarL v) = VarL v
 subsEE subs (Fn f arg) = Fn f $ map (subsEE subs) arg
 subsEE subs (Letrec env e) = Letrec (subsEE subs env) (subsEE subs e)
 subsEE subs (MetaExpression s e) = MetaExpression s e
 subsEE subs (Hole s) = (Hole s) 
 subsEE subs (MetaCtxt sub srt  c) = MetaCtxt sub srt c
 subsEE subs (MetaCtxtSubst sub srt c s) = (MetaCtxtSubst sub srt c (subsEE subs s))
instance SubsEE Binding where
 subsEE sub (x := e) = (x := subsEE sub e)
instance SubsEE ChainVar where
 subsEE subs (MetaChain sub typ ee x s)
  | Just ee' <- lookup ee subs = MetaChain sub typ ((case typ of {AChain -> "EE"; VarChain -> "VV"}) ++ ee') x s
  | otherwise = MetaChain sub typ ee x s
instance SubsEE Environment where
 subsEE subs env = 
  env {bindings = map (subsEE subs) (bindings env)
      ,metaChains = map (subsEE subs) (metaChains env)
     }
 
-- small helper function 
zipWithFresh op [] ys = ([],ys) 
zipWithFresh op (x:xs) (y:ys) = 
 let (rest,fresh') = (zipWithFresh op xs ys) 
 in ((x `op` y):rest,fresh')                      
                      
                      
                      
                      
                      
                      
                      
                      
                      
                      
                      
                      
-- =====================================================================================================================
-- =====================================================================================================================
-- =====================================================================================================================
-- =====================================================================================================================
-- =====================================================================================================================


       
       
applySol _ ((Node p _)) = error $ "PANIC: found open leaf " ++ show p       
-- for DEBUG only
applySol ((w1,n1,l1 :==>: r1), (w2,n2,l2 :==>: r2)) (Solved pb)
 | uexpr <- applySolution (solution pb) (inputGamma pb)
 , sigma1 <- solution pb
 , constraints <- delta pb
 , (any (\(x :=?=: y) -> x /= y) (expressionEq uexpr)) && nullGamma (delta4 constraints) = error "PANIC1a(applySol)"
 
applySol z@((w1,n1,l1 :==>: r1), (w2,n2,l2 :==>: r2)) xx@(Solved pb) 
 | uexpr <- applySolution (solution pb) (inputGamma pb)
 , sigma1 <- solution pb
 , constraints <- delta pb
 , ((applySolution sigma1 l1) /= (applySolution sigma1 l2)) && nullGamma (delta4 constraints) = 
 error $ "PANIC2(applySol)" ++  show z ++ "\n" ++ show xx
 ++ "\n" ++ show (applySolution sigma1 l1) ++ "\n" ++  show (applySolution sigma1 l2)
-- end of DEBUG 
applySol ((w1,n1,l1 :==>: r1), (w2,n2,l2 :==>: r2)) (Solved pb)
 | uexpr <- applySolution (solution pb) (inputGamma pb)
 , sigma1 <- solution pb
 , constraints <- delta pb
 , finalproblem <- pb 
 =
  let  sigma = sigma1 
       l1'   = applySolution sigma1 l1
       l2'   = applySolution sigma1 l2
       r1'   = applySolution sigma1 r1
       r2'   = applySolution sigma1 r2
       res   = ((w1,n1,l1' :==>: r1'), (w2,n2,l2' :==>: r2'), constraints,sigma1,(l1 :==>: r1,l2 :==>: r2),finalproblem)
  in 
--   DEBUG:  
--    case test of
--     Nothing -> 
     -- does not really work since lbeta can destroy the weak DVC, for instance:
     --
     --
    -- DVCFail r2: 
    -- (letrec {E21;X12=A22[A18[(letrec {X1=S16} in S7)]];X3=(\X1.S7)} in A19[(vlet X12)])
    -- Details:
    -- ((False,"rule_T_cp-e",
    -- (letrec {E21;X12=A22[A18[(app (\X1.S7) S16)]];X3=(\X1.S7)} in A19[(vlet X12)]) 
    -- ==> (letrec {E21;X12=A22[A18[(app (vlet X3) S16)]];X3=(\X1.S7)} in A19[(vlet X12)])),
    -- (True,"rule_NO_lbeta3",
    -- (letrec {E21;X12=A22[A18[(app (\X1.S7) S16)]];X3=(\X1.S7)} in A19[(vlet X12)])
    -- ==> (letrec {E21;X12=A22[A18[(letrec {X1=S16} in S7)]];X3=(\X1.S7)} in A19[(vlet X12)])),
    --
    --
    --     
     
     -- if  l1' /= l2'  && null (delta4 constraints) then error "PANIC4"
      -- else if null (delta4 constraints) then 
           -- if not $ checkWeakDVC l1' then error  ("DVCFail l1: " ++ show l1' ++"\n Details:\n" ++ show res ++ "\nsigma\n" ++ show sigma) else
           -- if not $ checkWeakDVC l2' then trace  ("DVCFail l2: " ++ show l2' ++"\n Details:\n" ++ show res ++ "\nsigma\n" ++ show sigma) else
           -- if not $ checkWeakDVC r1' then Nothing else
           -- if not $ checkWeakDVC r2' then Nothing else
           --                           Just res 
           -- else
--   END OF DEBUG    
         Just res
--     Just fail -> error $ "DOES THIS HAPPEN??? CAPCONFail" ++ show fail
-- --  where weaken sol = concat [  wk s    |  s <- sol]
-- --        wk     m@(MapVarLift ((MetaVariableLift x) :|-> r)) = [m, (MapVarLet ((MetaVariableLet x) :|-> wk1 r))]
-- --        wk     m@(MapVarLet ((MetaVariableLet x) :|-> r)) = [m, (MapVarLam ((MetaVarLift x) :|-> wk2 r))]
-- --        wk     r                      = [r]
-- --        wk1 (MetaVarLift x) = MetaVariableLet x
-- --        wk1 (VariableLam x)     = VariableLet x
-- --        wk2 (MetaVariableLet x) = MetaVarLift x
-- --        wk2 (VariableLet x)     = VariableLam x


       -- let (rules',fresh') = makeAllFresh rules fresh 
     -- problems        = map (\x -> execCommand x  rules') commands
     
-- makeProblem fnSyms (eqs,ctxtcon,envcon,capcon) frsh 
--   | trace (show eqs ++ "\n" ++ "\n" ++ show capcon) False = undefined
   
-- makeProblem finfo constrainedRule frsh = 
makeProblem fnSyms (eqs,ctxtcon,envcon,capcon) frsh = 
 let
  capconstr = capcon -- ++ concat [computeConstraints l ++ computeConstraints r     | (l:=?=:r) <- eqs]
  sol =[]
  gam =(Gamma {varLiftEq = []
              , expressionEq = eqs
              , environmentEq = []
              , bindingEq = []
              }) 
  del = (Delta {delta1=ctxtcon,delta2=envcon,delta3=capconstr,delta4=emptyGamma}) 
  fn = fnSyms
  prob = Problem {solution=sol,gamma=gam,delta=del,fresh=frsh,functions=fn,inputGamma=gam,inputDelta=del,constantDelta=emptyDelta,alphaBase=[],mpLogging = [],
                 expandAllContexts=False}
 in prob

-- =====================================================================================================================
-- =====================================================================================================================
-- =====================================================================================================================
 

 
 

 
