-- DONE: SYNTAX CHANGE

-----------------------------------------------------------------------------
-- | 
-- Module      :  LRSX.Language.Syntax
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--
--  Datatype for the internal representation 
--   of unification expressions
--
-----------------------------------------------------------------------------

-- todo: Work in VarPos everywhere

module LRSX.Language.Syntax where
import Data.List
import Debug.Trace
import Data.Function
import Data.Maybe
import LRSX.Util.Util
import qualified Data.Set as Set

-- ============================================================================
-- Some type synonyms
-- ============================================================================
-- Variables:
type Variable = String
-- Names of function symbols
type FName = String

-- ============================================================================
-- Syntax of unification expressions
-- ============================================================================
-- Idea: build the 'syntactic type system' into the data types,
-- i.e. for every type there is a Haskell data type, which 
-- already includes the Meta-variables

-- The two types of variables:
-- ===========================

data VarLift = 
    VarLift Substitution Variable 
  | MetaVarLift Substitution Variable
  | ConstantVarLift Substitution Variable -- only used for matching
 deriving(Eq)
 
instance Ord VarLift where
 compare (VarLift sub1 v1) (VarLift sub2 v2) = case compare v1 v2 of
                                                EQ -> compare sub1 sub2
                                                o -> o 
                                                
 compare (MetaVarLift sub1 v1) (MetaVarLift sub2 v2) = case compare v1 v2 of
                                                            EQ -> compare sub1 sub2
                                                            o -> o 

 compare (ConstantVarLift sub1 v1) (ConstantVarLift sub2 v2) = case compare v1 v2 of
                                                                EQ -> compare sub1 sub2
                                                                o -> o 
                                                
 compare (VarLift _ _) (MetaVarLift _ _ ) = LT
 compare (VarLift _ _) (ConstantVarLift _ _ ) = LT
 compare (MetaVarLift _ _) (VarLift _ _ )  = GT
 compare (MetaVarLift _ _) (ConstantVarLift _ _)  = LT
 compare (ConstantVarLift _ _) (VarLift _ _)  = GT
 compare (ConstantVarLift _ _) (MetaVarLift _ _)  = GT
 
 
-- Expressions:
-- ============
-- note that expressions also contain contexts, since
-- contexts can be viewed as expressions where the special
-- constant [.] (the hole) occurs.

type Context = Expression

-- Argument-Positions may have binders, or they are expressions (Binder [] expr), or they ar Variable-positions (VarPos X)
data ArgPos = Binder [VarLift] Expression | VarPos VarLift 
 deriving(Eq)
 
data Expression = 
    VarL VarLift                                              -- Lam-Variables lifted to expressions
  | Fn FName [ArgPos]                                         -- Function symbols, name, arguments, where arguments are expressions or binders
  | Letrec Environment Expression                             -- Letrec
  | MetaExpression Substitution String                        -- for meta variables representing expressions (S)
  | Hole Substitution                                         -- for contexts
  | MetaCtxt Substitution CtxtSort String                     -- for meta variables representing contexts     (C,T,A)
  | MetaCtxtSubst Substitution CtxtSort String Expression     -- ContextSubstitution, e.g. A[s] = MetaCtxtSubst ACtxt "A" s
  | ConstantCtxt Substitution CtxtSort String                 -- only used for matching
  | ConstantCtxtSubst Substitution CtxtSort String Expression -- only used for matching
  | Constant Substitution String                              -- Only used for matching
 deriving(Eq)
 
data Substitution = Substitution{substList::[SubstItem]}
 deriving(Eq,Ord) 
  
data Restriction a = None a | CapVars a | LetVars a  
 deriving (Eq,Ord)
instance Show a => Show (Restriction a) where
 show (None a) = show a
 show (CapVars a) = "CV(" ++ show a ++ ")"
 show (LetVars a) = "LV(" ++ show a ++ ")"
 
data SubstItem = SubstName (Restriction SyntaxComponent) String |  SubstVar VarLift VarLift | SubstSet [SubstItem]
instance Eq SubstItem where
    (SubstName a b) == (SubstName a' b') = a == a' && b == b'
    (SubstVar a b) == (SubstVar a' b') = a == a' && b == b'
    (SubstSet set1) == (SubstSet set2) = (Set.fromList set1) == (Set.fromList set2)
    _ == _ = False
instance Ord SubstItem where
    compare (SubstName _ _) (SubstVar _ _) = LT
    compare (SubstName _ _) (SubstSet _  ) = LT    
    compare (SubstName a b) (SubstName a' b') 
        | compare a a' == LT = LT
        | compare a a' == EQ = compare b b'
        | compare a a' == GT = GT
        
    compare (SubstVar a b)  (SubstVar a' b')  
        | compare a a' == LT = LT
        | compare a a' == EQ = compare b b'
        | compare a a' == GT = GT
    compare (SubstVar _ _) (SubstName _ _) = GT
    compare (SubstVar _ _) (SubstSet _ ) = LT
        
    compare (SubstSet set1) (SubstSet set2) = compare (Set.fromList set1) (Set.fromList set2)
    compare (SubstSet _) _ = GT    



    
noneToLetVars (SubstName (None a) b) = (SubstName (LetVars a) b)

 
 
instance Show Substitution where
  show sub 
    | null (substList sub) = ""
  show sub = "{[" ++ (intercalate "," (map show (substList sub))) ++ "]}"

instance Show SubstItem where
  show (SubstName comp name) = name ++ "_" ++ show comp
  show (SubstVar l r) = show l ++ "->" ++ show r
  show (SubstSet set) = "{" ++ (intercalate "," (map show set))  ++ "}"
  
isEmptySubstitution sub =  null (substList sub)  
emptySubstitution = Substitution {substList = []}
hole = Hole emptySubstitution                            
-- Bindings:
-- =========
data Binding = VarLift := Expression
 deriving(Eq)

-- Environments:
-- ============
-- an environment has three parts: 
--  - a list of bindings
--  - a list of meta-environments
--  - a list of meta-chains
 
data Environment = 
 Environment { bindings         :: [Binding],
               metaEnvVars      :: [EnvVar],
               metaChains       :: [ChainVar],
               constantEnvVars  :: [EnvVar], -- only used for matching
               constantChains   :: [ChainVar] -- only used for matching
              }
              
data EnvVar = Env Substitution String 
 deriving(Eq)
 
instance Ord EnvVar where
  compare (Env sub1 x1) (Env sub2 x2) 
    | EQ <- compare x1 x2 = compare sub1 sub2
    | otherwise = compare x1 x2
    
instance Show EnvVar where
  show (Env subst s)                                   = s ++ show subst

-- Chain Variables
-- =============== 
data ChainVar = MetaChain Substitution ChainType String VarLift Expression -- EE[x,s] == MetaChain "EE" x s
              | ConstantChain Substitution ChainType String VarLift Expression -- only used for matching
 deriving(Eq)
 
data ChainType = AChain | VarChain
 deriving(Eq,Show)
 
isMetaCtxtSubst (MetaCtxtSubst _ _ _ _) = True
isMetaCtxtSubst _ = False
-- ============================================================================= 
-- Ord-instances 
instance Ord ArgPos where
 compare (VarPos _) (Binder _ _) = LT
 compare (Binder _ _) (VarPos _) = GT
 compare (VarPos v1) (VarPos v2) = compare v1 v2  
 compare (Binder v e) (Binder v' e') 
   | EQ <- compare v v'         = compare e e'
   | otherwise                  = compare v v'
   
instance Ord Expression where
 compare (VarL x) (VarL y)   = compare x y
 compare (VarL x) _          = LT
   
 compare (Fn f args) (Fn g args')
  | EQ <- compare f g            = compare args args'
  | otherwise                    = compare f g
 compare (Fn _ _)  (VarL _)    = GT
 compare (Fn _ _)  _             = LT
 
 compare (Letrec env1 e1) (Letrec env2 e2)
  | EQ <- compare e1 e2           = compare env1 env2
  | otherwise                     = compare e1 e2
 compare (Letrec _ _) (VarL _)  = GT
 compare (Letrec _ _) (Fn _ _)    = GT
 compare (Letrec _ _) _           = LT
 
 compare (MetaExpression sub1 s) (MetaExpression sub2 s') = case compare s s' of 
                                                               EQ -> compare sub1 sub2
                                                               o -> o
 compare (MetaExpression _ _) (VarL _)  = GT
 compare (MetaExpression _ _) (Fn _ _)    = GT
 compare (MetaExpression _ _) (Letrec _ _)    = GT
 compare (MetaExpression _ _) _           = LT
 
 compare (Hole sub1) (Hole sub2) = compare sub1 sub2
 compare (Hole _) (VarL _)  = GT
 compare (Hole _) (Fn _ _)    = GT
 compare (Hole _) (Letrec _ _)    = GT
 compare (Hole _) (MetaExpression _ _) = GT
 compare (Hole _) _           = LT
 
 compare (MetaCtxt sub1 srt1 n1) (MetaCtxt sub2 srt2 n2)
   | EQ <- compare n1 n2 = case compare srt1 srt2 of 
                             EQ -> compare sub1 sub2 
                             o -> o
   | otherwise = compare n1 n2
 compare (MetaCtxt sub1 srt1 name1) (MetaCtxtSubst sub2 srt2 name2 (Hole _))
   | srt1 == srt2 && name1 == name2 = EQ
   
 compare (MetaCtxt _ _ _) (VarL _)  = GT
 compare (MetaCtxt _ _ _) (Fn _ _)    = GT
 compare (MetaCtxt _ _ _) (Letrec _ _)    = GT
 compare (MetaCtxt _ _ _) (MetaExpression _ _) = GT
 compare (MetaCtxt _ _ _) (Hole _) = GT
 compare (MetaCtxt _ _ _) _           = LT
 
 compare (MetaCtxtSubst sub1 srt1 n1 e1) (MetaCtxtSubst sub2 srt2 n2 e2)
   | EQ <- compare n1 n2 = case compare e1 e2 of
                             EQ -> compare (MetaCtxt sub1 srt1 n1) (MetaCtxt sub2 srt2 n2)
                             other -> other
   | otherwise = compare n1 n2
 compare (MetaCtxtSubst _ _ _ _) (VarL _)  = GT
 compare (MetaCtxtSubst _ _ _ _) (Fn _ _)    = GT
 compare (MetaCtxtSubst _ _ _ _) (Letrec _ _)    = GT
 compare (MetaCtxtSubst _ _ _ _) (MetaExpression _ _) = GT
 compare (MetaCtxtSubst _ _ _ _) (Hole _) = GT
 compare (MetaCtxtSubst _ srt1 name1 (Hole _)) (MetaCtxt _ srt2 name2)
   | srt1 == srt2 && name1 == name2 = EQ
 
 compare (MetaCtxtSubst _ _ _ _) (MetaCtxt _ _ _) = GT
 compare (MetaCtxtSubst _ _ _ _) _           = LT

 compare (ConstantCtxtSubst sub1 srt1 n1 e1) (ConstantCtxtSubst sub2 srt2 n2 e2)
   | EQ <- compare n1 n2 = case compare e1 e2 of
                             EQ -> compare (ConstantCtxt sub1 srt1 n1) (ConstantCtxt sub2 srt2 n2)
                             other -> other
   | otherwise = compare n1 n2
   
 compare (ConstantCtxtSubst sub1 srt1 name1 (Hole _)) (ConstantCtxt sub2 srt2 name2)
   | srt1 == srt2 && name1 == name2 && sub1 == sub2 = EQ
 compare (ConstantCtxtSubst _ _ _ _) (VarL _)  = GT
 compare (ConstantCtxtSubst _ _ _ _) (Fn _ _)    = GT
 compare (ConstantCtxtSubst _ _ _ _) (Letrec _ _)    = GT
 compare (ConstantCtxtSubst _ _ _ _) (MetaExpression _ _) = GT
 compare (ConstantCtxtSubst _ _ _ _) (Hole _) = GT
 compare (ConstantCtxtSubst _ _ _ _) (MetaCtxt _ _ _) = GT
 compare (ConstantCtxtSubst _ _ _ _) (MetaCtxtSubst _ _ _ _) = GT
 compare (ConstantCtxtSubst _ _ _ _) _           = LT

 compare (ConstantCtxt sub1 srt1 n1 ) (ConstantCtxt sub2 srt2 n2)
   | EQ <- compare n1 n2 = case compare srt1 srt2 of 
                             EQ -> compare sub1 sub2 
                             o -> o
   | otherwise = compare n1 n2
   
 compare (ConstantCtxt _ _ _) (VarL _)  = GT
 compare (ConstantCtxt _ _ _) (Fn _ _)    = GT
 compare (ConstantCtxt _ _ _ ) (Letrec _ _)    = GT
 compare (ConstantCtxt _ _ _) (MetaExpression _ _) = GT
 compare (ConstantCtxt _ _ _) (Hole _) = GT
 compare (ConstantCtxt _ _ _) (MetaCtxt _ _ _) = GT
 compare (ConstantCtxt _ _ _) (MetaCtxtSubst _ _ _ _) = GT
 compare (ConstantCtxt _ srt1 name1) (ConstantCtxtSubst _ srt2 name2 (Hole _))
   | srt1 == srt2 && name1 == name2 = EQ
 compare (ConstantCtxt _ _ _) (ConstantCtxtSubst _ _ _ _) = GT
 compare (ConstantCtxt _ _ _) _           = LT

 compare (Constant sub1 s) (Constant sub2 s') = case compare s s' of
                                                  EQ -> compare sub1 sub2
                                                  o -> o
 compare (Constant _ s) _ = GT
 
 
 
 compare a b = error $ show a ++ show b
 
 
instance Ord Binding where
 compare (x1:=e1) (x2:=e2)
  | EQ <- compare x1 x2 = compare e1 e2
  | otherwise = compare x1 x2  
   
instance Ord Environment where 
 compare env1 env2 =
   let c1 = compare  
             (sort (bindings env1))
             (sort (bindings env2))
       c2 = compare (sort $ metaEnvVars env1) (sort $ metaEnvVars env2)
       c3 = compare (sort (metaChains env1)) (sort  (metaChains env2))
       c4 = compare (sort $ constantEnvVars env1) (sort $ constantEnvVars env2)
       c5 = compare (sort (constantChains env1)) (sort (constantChains env2))
   in case c1 of 
       EQ -> case c2 of 
               EQ -> case c3 of 
                       EQ -> case c4 of 
                               EQ -> c5
                               other -> other
                       other -> other
               other -> other
       other -> other               
  where
     bcomp (x := e) (y := e')  
       | EQ <- compare x y = compare e e'
       | otherwise = compare x y

instance Ord ChainVar where
 compare (MetaChain sub1 typ1 ee1 z1 s1) (MetaChain sub2 typ2 ee2 z2 s2) =
   case compare ee1 ee2 of 
     EQ -> case compare typ1 typ2 of
             EQ -> case compare z1 z2 of
                     EQ -> case compare s1 s2 of
                             EQ -> compare sub1 sub2
                             o -> o
                     other -> other
             other -> other
     other -> other
 compare (MetaChain _ typ1 ee1 z1 s1) _ = LT
 
 compare (ConstantChain sub1 typ1 ee1 z1 s1) (ConstantChain sub2 typ2 ee2 z2 s2) =
   case compare ee1 ee2 of 
     EQ -> case compare typ1 typ2 of
             EQ -> case compare z1 z2 of
                     EQ -> case compare s1 s2 of
                              EQ -> compare sub1 sub2
                              o -> o
                     other -> other
             other -> other
     other -> other
 compare (ConstantChain _ _ _ _ _) _ = GT
 
instance Ord ChainType where
 VarChain <= _ = True
 _        <= AChain   = True
 _        <= _ = False

instance Eq Environment where
 env1 == env2 = compare env1 env2 == EQ

-- data type for syntactic components 
data SyntaxComponent =
     SCVar VarLift
  |  SCExpr Expression
  |  SCCtxt Expression
  |  SCMetaEnv EnvVar
  |  SCMetaChain Substitution ChainType String  
  |  SCConstantEnv EnvVar
  |  SCConstantChain Substitution ChainType String  
 deriving(Eq)
 
instance Ord SyntaxComponent where
 compare (SCVar v1) (SCVar v2) = compare v1 v2
 compare (SCVar _) _ = LT

 compare (SCExpr e1) (SCVar  _ ) = GT
 compare (SCExpr e1) (SCExpr e2) = compare e1 e2
 compare (SCExpr e1) _ = LT

 compare (SCCtxt _) (SCVar _) = GT
 compare (SCCtxt _) (SCExpr _) = GT
 compare (SCCtxt c1) (SCCtxt c2) = compare c1 c2 
 compare (SCCtxt _) _ = LT
 
 compare (SCMetaEnv _) (SCVar _) = GT
 compare (SCMetaEnv _) (SCExpr _) = GT
 compare (SCMetaEnv _) (SCCtxt _) = GT
 compare (SCMetaEnv e1) (SCMetaEnv e2) = compare e1 e2
 compare (SCMetaEnv e1) _ = LT 

 compare (SCMetaChain _ _ _) (SCVar _) = GT
 compare (SCMetaChain _ _ _) (SCExpr _) = GT
 compare (SCMetaChain _ _ _) (SCCtxt _) = GT
 compare (SCMetaChain _ _ _) (SCMetaEnv _) = GT
 compare (SCMetaChain sub t s) (SCMetaChain sub' t' s') 
  | EQ <- compare s s' = case compare t t' of
                          EQ -> compare sub sub'
                          oter -> oter
  | otherwise = compare s s'  
 compare (SCMetaChain sub t s) _ = LT
 
 compare (SCConstantChain _ _ _) (SCVar _) = GT
 compare (SCConstantChain _ _ _) (SCExpr _) = GT
 compare (SCConstantChain _ _ _) (SCCtxt _) = GT
 compare (SCConstantChain _ _ _) (SCMetaEnv _) = GT
 compare (SCConstantChain _ _ _) (SCMetaChain _ _ _) = GT
 compare (SCConstantChain sub t s) (SCConstantChain sub' t' s') 
  | EQ <- compare s s' = case compare t t' of
                          EQ -> compare sub sub'
                          oter -> oter
  | otherwise = compare s s'  
 compare (SCConstantChain sub t s) _ = LT
 
 compare (SCConstantEnv _ ) (SCVar _) = GT
 compare (SCConstantEnv _ ) (SCExpr _) = GT
 compare (SCConstantEnv _ ) (SCCtxt _) = GT
 compare (SCConstantEnv _ ) (SCMetaEnv _) = GT
 compare (SCConstantEnv _ ) (SCMetaChain _ _ _) = GT
 compare (SCConstantEnv _ ) (SCConstantChain _ _ _) = GT
 compare (SCConstantEnv e1 ) (SCConstantEnv e2 ) = compare e1 e2
 
instance Show SyntaxComponent where
 show (SCVar v) = show v
 show (SCExpr e) = show e
 show (SCCtxt d ) = show d
 show (SCMetaEnv e) = show e
 show (SCConstantEnv (Env sub name)) = name ++ "*" ++ show sub
 show (SCMetaChain sub typ string) = string ++ show sub
 show (SCConstantChain sub typ string) = string ++ "*" ++ show sub

 
isConcreteSCVar (SCVar (VarLift (Substitution []) _)) = True
isConcreteSCVar (SCVar (ConstantVarLift (Substitution []) _)) = True
isConcreteSCVar _ = False
-- type class to compute syntact components
class SyntaxComponents a where
 -- compute syntactic components that may have bound variables
 getSCBoundVars :: a -> Set.Set SyntaxComponent
 -- compute syntactic components that may have free or bound variables
 getSCAllVars :: a -> Set.Set SyntaxComponent
 -- compute syntactic components that may have binders that affect the hole of a context (compute CV(.))
 getSCCapVars :: a -> Set.Set SyntaxComponent

instance SyntaxComponents VarLift where
 getSCBoundVars v = Set.empty
 getSCAllVars   v = Set.singleton (SCVar v)
 getSCCapVars   v = Set.empty
 

instance SyntaxComponents ArgPos where
 getSCBoundVars (VarPos v)    = Set.empty
 getSCBoundVars (Binder vs e) = (Set.unions (map getSCAllVars vs)) `Set.union` (getSCBoundVars e)
 getSCAllVars   (VarPos v)    = Set.singleton (SCVar v)
 getSCAllVars   (Binder vs e) = (Set.unions (map getSCAllVars vs)) `Set.union` (getSCAllVars e)
 getSCCapVars   (VarPos v)    = Set.empty
 getSCCapVars   (Binder vs e)
   | hasHole e                = (Set.unions (map getSCAllVars vs)) `Set.union` (getSCCapVars e)
   | otherwise                = (getSCCapVars e)
 
instance SyntaxComponents Expression where
 getSCBoundVars (VarL v)        = getSCBoundVars v 
 getSCBoundVars (Fn f args)     = (Set.unions (map getSCBoundVars args))
 getSCBoundVars (Letrec env e)  = getSCBoundVars e `Set.union` getSCBoundVars env
 getSCBoundVars m@(MetaExpression sub s) = Set.singleton $ SCExpr m 
 getSCBoundVars m@(Constant sub s)       = Set.singleton $ SCExpr m 
 getSCBoundVars m@(MetaCtxt sub _ _)     = Set.singleton $ SCCtxt m
 getSCBoundVars m@(ConstantCtxt sub _ _) = Set.singleton $ SCCtxt m
 getSCBoundVars m@(MetaCtxtSubst sub srt name expr) = 
   (Set.singleton $ SCCtxt (MetaCtxt sub srt name))
   `Set.union`
   (getSCBoundVars expr)
 getSCBoundVars m@(ConstantCtxtSubst sub srt name expr) = 
   (Set.singleton $ SCCtxt (ConstantCtxt sub srt name))
   `Set.union`
   (getSCBoundVars expr)
 getSCBoundVars (Hole _) = Set.empty

 getSCAllVars (VarL v)          = getSCAllVars v 
 getSCAllVars (Fn f args)     = (Set.unions (map getSCAllVars args))
 getSCAllVars (Letrec env e)  = getSCAllVars e `Set.union` getSCAllVars env
 getSCAllVars m@(MetaExpression sub s) = Set.singleton $ SCExpr m 
 getSCAllVars m@(Constant sub s)       = Set.singleton $ SCExpr m 
 getSCAllVars m@(MetaCtxt sub _ _)     = Set.singleton $ SCCtxt m
 getSCAllVars m@(ConstantCtxt sub _ _) = Set.singleton $ SCCtxt m
 getSCAllVars m@(MetaCtxtSubst sub srt name expr) = 
   (Set.singleton $ SCCtxt (MetaCtxt sub srt name))
   `Set.union`
   (getSCAllVars expr)
 getSCAllVars m@(ConstantCtxtSubst sub srt name expr) = 
   (Set.singleton $ SCCtxt (ConstantCtxt sub srt name))
   `Set.union`
   (getSCAllVars expr)
 getSCAllVars (Hole _) = Set.empty

 getSCCapVars (VarL v)          = getSCCapVars v 
 getSCCapVars (Fn f args)       = (Set.unions (map getSCCapVars args))
 getSCCapVars (Letrec env e)    
   | hasHole e   = 
          Set.unions
           [Set.unions   [ (getSCAllVars z) `Set.union` (Set.singleton $ SCMetaChain sub typ name) |  (MetaChain sub typ name z s) <- metaChains env]
           ,Set.unions   [ (getSCAllVars z) `Set.union` (Set.singleton $ SCConstantChain sub typ name) |  (ConstantChain sub typ name z s) <- constantChains env]
           ,Set.fromList [ (SCMetaEnv envvar) | envvar <- metaEnvVars env]
           ,Set.fromList [ (SCConstantEnv envvar) | envvar <- constantEnvVars env]
           ,Set.unions   [ (getSCAllVars z) | (z:= s) <- bindings env]
           ,getSCCapVars e
           ] 
   | otherwise = getSCCapVars env
 getSCCapVars   m@(MetaExpression sub s) = Set.empty 
 getSCCapVars m@(Constant sub s)       = Set.empty 
 getSCCapVars m@(MetaCtxt sub srt _)     = 
   if srt  `elem` [TCtxt,Ctxt] then 
    Set.singleton $ SCCtxt m
    else Set.empty
 getSCCapVars m@(ConstantCtxt sub srt _) = 
   if srt `elem` [TCtxt,Ctxt] then 
    Set.singleton $ SCCtxt m
    else Set.empty
 getSCCapVars m@(MetaCtxtSubst sub srt name expr) 
  | hasHole expr = (if srt `elem` [Ctxt,TCtxt]
                     then
                         Set.singleton $ SCCtxt (MetaCtxt sub srt name)
                     else Set.empty
                   ) `Set.union` (getSCCapVars expr)
  | otherwise = Set.empty
 getSCCapVars m@(ConstantCtxtSubst sub srt name expr) 
  | hasHole expr = (if srt `elem` [Ctxt,TCtxt]
                     then 
                          Set.singleton $ SCCtxt (ConstantCtxt sub srt name)
                     else Set.empty
                   ) `Set.union` (getSCCapVars expr)
  | otherwise = Set.empty
 getSCCapVars (Hole _) = Set.empty
 
instance SyntaxComponents Environment where
 getSCBoundVars env =
  Set.unions
   [Set.unions   [ (getSCAllVars z) `Set.union` (getSCBoundVars s) `Set.union` (Set.singleton $  SCMetaChain sub typ name)
                    |  (MetaChain sub typ name z s) <- metaChains env
                 ]
   ,Set.unions   [ (getSCAllVars z) `Set.union` (getSCBoundVars s) `Set.union` (Set.singleton $  SCConstantChain sub typ name)
                    |  (ConstantChain sub typ name z s) <- constantChains env
                 ]
   ,Set.fromList [ (SCMetaEnv envvar) | envvar <- metaEnvVars env]
   ,Set.fromList [ (SCConstantEnv envvar) | envvar <- constantEnvVars env]
   ,Set.unions   [(getSCAllVars z) `Set.union` (getSCBoundVars s) | (z:= s) <- bindings env]
   ]
 getSCAllVars env =
  Set.unions
   [Set.unions   [ (getSCAllVars z) `Set.union` (getSCAllVars s) `Set.union` (Set.singleton $  SCMetaChain sub typ name)
                    |  (MetaChain sub typ name z s) <- metaChains env
                 ]
   ,Set.unions   [ (getSCAllVars z) `Set.union` (getSCAllVars s) `Set.union` (Set.singleton $  SCConstantChain sub typ name)
                    |  (ConstantChain sub typ name z s) <- constantChains env
                 ]
   ,Set.fromList [ (SCMetaEnv envvar) | envvar <- metaEnvVars env]
   ,Set.fromList [ (SCConstantEnv envvar) | envvar <- constantEnvVars env]
   ,Set.unions   [(getSCAllVars z) `Set.union` (getSCAllVars s) | (z:= s) <- bindings env]
   ]
 getSCCapVars env =
  let binders = 
          Set.unions
           [Set.unions   [ (getSCAllVars z) `Set.union` (Set.singleton $ SCMetaChain sub typ name) |  (MetaChain sub typ name z s) <- metaChains env]
           ,Set.unions   [ (getSCAllVars z) `Set.union` (Set.singleton $ SCConstantChain sub typ name) |  (ConstantChain sub typ name z s) <- constantChains env]
           ,Set.fromList [ (SCMetaEnv envvar) | envvar <- metaEnvVars env]
           ,Set.fromList [ (SCConstantEnv envvar) | envvar <- constantEnvVars env]
           ,Set.unions   [ (getSCAllVars z) | (z:= s) <- bindings env]
           ]
      exprWithHole =
          Set.unions
           [Set.unions   [ (getSCCapVars s)   | (MetaChain sub typ name z s) <- metaChains env, hasHole s]
           ,Set.unions   [ (getSCCapVars s)    | (ConstantChain sub typ name z s) <- constantChains env, hasHole s]
           ,Set.unions   [ (getSCCapVars s) | (z:= s) <- bindings env, hasHole s]
           ]
      noHole   = null
          (concat 
          [   [ ()   | (MetaChain sub typ name z s) <- metaChains env, hasHole s]
           ,  [ ()   | (ConstantChain sub typ name z s) <- constantChains env, hasHole s]
           ,  [ ()   | (z:= s) <- bindings env, hasHole s]
           ])

  in 
      
      if noHole then 
     
        Set.empty else binders `Set.union` exprWithHole


instance (SyntaxComponents a, SyntaxComponents b) => SyntaxComponents (Either a b) where
 getSCBoundVars (Left a)  = getSCBoundVars a
 getSCBoundVars (Right a) = getSCBoundVars a
 getSCAllVars (Left a)    = getSCAllVars a
 getSCAllVars (Right a)   = getSCAllVars a
 getSCCapVars (Left a)    = getSCCapVars a
 getSCCapVars (Right a)   = getSCCapVars a
   
   

-- fresh symbols are meta variables: 
isFreshSym (SCVar (MetaVarLift _ _)) = True
isFreshSym (SCExpr (MetaExpression _ _)) = True
isFreshSym (SCCtxt (MetaCtxt _ _ _)) = True
isFreshSym (SCMetaEnv _) = True
isFreshSym (SCMetaChain _ _ _) = True
isFreshSym _ = False
 
-- =============================================================================
 
-- several helper functions:
-- create an empty environment
empty :: Environment
empty = Environment {bindings =[], metaEnvVars =[], metaChains =[],constantChains =[], constantEnvVars = []}
-- create Environment from a single binding:
binding :: Binding -> Environment
binding b = empty{bindings = [b]}
-- create Environment from a meta chain
metaChain :: ChainType -> String -> VarLift -> Expression -> Environment
metaChain typ ee v e = empty{metaChains = [MetaChain emptySubstitution typ ee v e]}
-- create Environment containing only a Meta-environment
metaEnv :: EnvVar -> Environment
metaEnv var = empty{metaEnvVars = [var]}
-- check whether an Environment is only a meta-environment
isMetaEnv :: Environment -> Bool
isMetaEnv env = null (bindings env) && null (metaChains env) && null (constantChains env) && null (constantEnvVars env) && isSingleton (metaEnvVars env)
-- check whether an Environment contains a binding
hasBinding :: Environment -> Bool
hasBinding env = not (null (bindings env))
-- check whether an Environment contains a meta-chain
hasMetaChain :: Environment -> Bool
hasMetaChain env = not (null (metaChains env))
-- check whether an Environment is empty
isEmpty :: Environment -> Bool
isEmpty env = and [null (bindings env), null (metaEnvVars env), null (metaChains env), null (constantChains env), null (constantEnvVars env)]

 

-- Context-Sorts
-- =============
data CtxtSort = ACtxt | TCtxt | Ctxt  
 deriving(Eq,Show)
 
instance Ord CtxtSort where
 ACtxt <= _     = True
 _     <= Ctxt  = True
 TCtxt <= TCtxt = True
 _ <= _ = False

repr :: CtxtSort -> String
repr ACtxt = "A"
repr TCtxt = "T"
repr Ctxt  = "C"

-- =======================================
-- class for computing meta-contexts
 
class HasMetaContexts a where
 computeMetaContexts :: a -> [Expression] 
  
instance HasMetaContexts ArgPos where
 computeMetaContexts (Binder v e)  = computeMetaContexts e
 computeMetaContexts (VarPos _) = []
 
instance HasMetaContexts Expression where
 computeMetaContexts (VarL v)  = []
 computeMetaContexts (Fn f arg) =  concatMap (computeMetaContexts) arg
 computeMetaContexts (Letrec env e) = (computeMetaContexts env) ++ (computeMetaContexts e)
 computeMetaContexts (MetaExpression _ e) = []
 computeMetaContexts (Hole _) = []
 computeMetaContexts (MetaCtxt subst srt  c) = [MetaCtxt subst srt c]
 computeMetaContexts (MetaCtxtSubst subst srt c s) = [MetaCtxt subst srt c] ++ computeMetaContexts s
 computeMetaContexts (Constant _ e) = []
 computeMetaContexts (ConstantCtxt _ srt  c) = [] -- ?
 computeMetaContexts (ConstantCtxtSubst _ srt c s) =  computeMetaContexts s -- ?
instance HasMetaContexts Binding where
 computeMetaContexts (x := e) = (computeMetaContexts e) 
instance HasMetaContexts ChainVar where
 computeMetaContexts (MetaChain _ _ ee x s) = computeMetaContexts s
instance HasMetaContexts Environment where
 computeMetaContexts env = 
  (concatMap computeMetaContexts (bindings env))
  ++ (concatMap computeMetaContexts (metaChains env))
  ++ (concatMap computeMetaContexts (constantChains env))
 
-- ============================================================================


-- ============================================================================
-- Show-instances
-- ============================================================================
 
instance Show ArgPos where 
 show (Binder [] e) = show e
 show (Binder v e) =  (intercalate "." (map show v)) ++ "." ++ show e 
 show (VarPos v) = show v

instance Show Expression where
 show (VarL v) = "(var " ++ show v ++ ")"
 show (Fn f args) = "(" ++ f   ++ " " ++ (concat (intersperse " " (map show args))) ++ ")"
 -- show (FnExt f spos args) = "(" ++ f   ++ " " ++ show spos ++  (concat (intersperse " " (map show args))) ++ ")"
 show (Letrec env e) = "(letrec " ++ show env ++ " in " ++ show e ++ ")"
 show (MetaExpression subst e) = e ++ show subst
 show (Hole subst) = "[.]" ++ show subst
 show (MetaCtxt sub srt  c) = c   ++ show sub -- ++ "_" ++ show srt
 show (MetaCtxtSubst sub srt c s) = c ++ show sub -- ++ "_" ++ show srt  
                                  ++ "[" ++ show s ++ "]"
 show (Constant subst e) = e ++ "*" ++ show subst
 show (ConstantCtxt subst srt  c) = c ++ "*" ++ show subst -- ++ "_" ++ show srt
 show (ConstantCtxtSubst subst srt c s) = c++ "*" ++ show subst  -- ++ "_" ++ show srt  
                                  ++ "[" ++ show s ++ "]"
instance Show VarLift where 
 show (VarLift subst v) = v  ++ show subst -- ++ "vlet"
 show (MetaVarLift subst v) = v ++ show subst-- ++ "vlet"
 show (ConstantVarLift subst v) = v ++ "*" ++ show subst

instance Show Binding where
 show (a := b) = show a ++ "=" ++ show b 

instance Show ChainVar where
  show (MetaChain subst AChain s v e)       = s ++ show subst ++ "[" ++ show v ++ "," ++ show e ++ "]"
  show (MetaChain subst VarChain s v e)     = s ++ show subst ++ "|" ++ show v ++ "," ++ show e ++ "|"
  show (ConstantChain subst AChain s v e)   = s ++ show subst ++ "*[" ++ show v ++ "," ++ show e ++ "]"
  show (ConstantChain subst VarChain s v e) = s ++ show subst ++ "*|" ++ show v ++ "," ++ show e ++ "|"
 
instance Show Environment where
 show env =
  let bs = map show $ sort $ bindings env
      ms = map show $ sort $ metaEnvVars env
      mss = map (\x -> "*" ++ x) $ map show $ sort $ constantEnvVars env
      mc = map show $ sort $ metaChains env
      mmc =  map show $ sort $ constantChains env
      all =  bs ++ ms ++ mc ++ mss ++ mmc
--       all =  ["bindings:"]++ bs ++ 
--              ["metaEnvV:"] ++ ms ++
--              ["metaChains:"] ++mc ++ 
--              ["constantEnvV:"] ++ mss  ++
--              ["constantChains:"] ++ mmc      
   in if null all then "{}" else "{" ++ (concat $ intersperse ";" all) ++ "}"
   
--
 
-- ============================================================================

-- ============================================================================
-- Instances for ConvExpression
-- ============================================================================

  
instance ConvExpression VarLift where
  toExpression var = VarL var
  fromExpression (VarL var) = var
  fromExpression e = MetaVarLift emptySubstitution "__hidden__" -- this of course is a hack
   
instance ConvExpression Expression where
  toExpression = id
  fromExpression = id  
-- ============================================================================



-- ============================================================================
-- Instances for the substitution
-- ============================================================================

instance Substitute VarLift where
 subst e1 e2 = substVarL (fromExpression $ toExpression e1) (fromExpression $ toExpression e2)
  where
    substVarL vary (MetaVarLift _ x) (MetaVarLift _ z)
      | x == z = vary
    substVarL     _ _ e = e 
 substEmpty  _ e = e
 substEnv  _ _ e = e 
 substChainVar _ _ e = e
 substEnvConstraint  _ _ e = e 
--  


instance Substitute ArgPos where
 subst e1 e2 b = 
    substIt (toExpression e1) (toExpression e2) b
  where 
   substIt e1@(VarL vary) e2@(VarL (MetaVarLift _ x)) (VarPos (MetaVarLift _ z)) 
    | x == z = VarPos vary
   substIt e1 e2 (VarPos v) = VarPos v
   substIt e1@(VarL vary) e2@(VarL varx) (Binder vars e) = Binder (map (subst vary varx) vars) (subst e1 e2 e)
   substIt e1 e2 (Binder v e)              = Binder v (subst e1 e2 e)
 substEmpty n (VarPos v)               = VarPos v
 substEmpty n (Binder v e)             = Binder v (substEmpty n e)
 substEnv env n (VarPos v )            = VarPos v
 substEnv env n (Binder v e)           = Binder v (substEnv env n e)
 substChainVar env n (VarPos v)        = VarPos v
 substChainVar env n (Binder v e)      = Binder v (substChainVar env n e)
 substEnvConstraint env n (VarPos v )  = VarPos v
 substEnvConstraint env n (Binder v e) = Binder v (substEnvConstraint env n e)
 
instance Substitute Expression where
 subst e1 e2 = substExpression (toExpression e1) (toExpression e2)
  where
   -- usual cases, where expressions are substituted by expressions:
   substExpression e1 (VarL varx) (VarL varz)
    | varx == varz = e1
    | otherwise = (VarL varz)
   substExpression e1 e2            (VarL varz) = VarL varz
    

   substExpression e1 e2 (Fn name args) = Fn name (map (subst e1 e2) args)

   substExpression e1 e2 (Letrec env inexpr) = Letrec (subst e1 e2 env) (substExpression e1 e2 inexpr)    

   substExpression e1 (MetaExpression subst1 s1) (MetaExpression subst2 s2)
    | s1 == s2 && subst1 == subst2 = e1
   substExpression e1 e2 (MetaExpression subst s2) = MetaExpression subst s2
   substExpression e1 e2 (Constant subst s) = Constant subst s -- do not substitute into constants

   -- substitution of a hole by an expression or a context
   substExpression e1 (Hole _) (Hole _) = e1
   substExpression e1 e2 (Hole e) = (Hole e)

   -- substitution of a context symbol by a context
   substExpression e1 (MetaCtxt _ s name) (MetaCtxt _ s' name2)
    | name == name2 = e1
   -- substitution of the hole of a meta-ctxt, i.e. C[ctxt/[.]] ---> C[ctxt] 
   substExpression context (Hole _) (MetaCtxt subst s name) = MetaCtxtSubst subst s name  context 
   substExpression context (Hole _) (ConstantCtxt subst s name) = ConstantCtxtSubst subst s name  context 
   -- substitution of an expression by an expression in a meta-ctxt has no effect:   
   substExpression e1 e2 (MetaCtxt subst s name2) = MetaCtxt subst s name2
   substExpression e1 e2 (ConstantCtxt subst s name2) = ConstantCtxt subst s name2
   -- substitution of a meta-ctxt by another meta-ctxt in sth like C[s], i.e. C[s][D/C] = D[s]
   substExpression e1@(MetaCtxt subst1 s0 new) e2@(MetaCtxt subst2 s name) (MetaCtxtSubst subst3 s' name2 e) 
    | name == name2 && subst2 ==subst3 = MetaCtxtSubst subst1 s0 new (substExpression e1 e2 e)                    
   -- substitution of a meta-ctxt by another context in sth like C[s], i.e. C[s][ctxt/C]=ctxt[s] = ctxt[.][s/[.]
   substExpression context e2@(MetaCtxt subst1 s name) (MetaCtxtSubst subst2 s' name2 e)  
    | name == name2 && subst1 == subst2 = subst e (Hole emptySubstitution) context                                
   -- substitution e1/e2 in sth like C[s] => C[s[e1/e2]]    
   substExpression e1 e2 (MetaCtxtSubst subst s name2 e) = (MetaCtxtSubst subst s name2 (substExpression e1 e2 e))
   substExpression e1 e2 (ConstantCtxtSubst subst s name2 e) = (ConstantCtxtSubst subst s name2 (substExpression e1 e2 e))

-- substEmpty n (Lam v e) = Lam v (substEmpty n e)
 substEmpty n (Fn name args) = Fn name (map (substEmpty n) args)
 substEmpty n (MetaCtxtSubst subst c name e) = MetaCtxtSubst subst c name (substEmpty n e) 
 substEmpty n (ConstantCtxtSubst subst c name e) = ConstantCtxtSubst subst c name (substEmpty n e) 
 substEmpty n (Letrec env inexpr) = Letrec (substEmpty n env) (substEmpty n inexpr)    
 substEmpty n e = e 

-- substEnv env n (Lam v e) = Lam v (substEnv env n e)
 substEnv env n (Fn name args) = Fn name (map (substEnv env n) args)
 substEnv env n (MetaCtxtSubst s c name e) = MetaCtxtSubst s c name (substEnv env n e) 
 substEnv env n (ConstantCtxtSubst s c name e) = ConstantCtxtSubst s c name (substEnv env n e) 
 substEnv env n (Letrec env' inexpr) = Letrec (substEnv env n env') (substEnv env n inexpr)    
 substEnv env n e = e 

-- substChainVar env n (Lam v e) = Lam v (substChainVar env n e)
 substChainVar env n (Fn name args) = Fn name (map (substChainVar env n) args)
 substChainVar env n (MetaCtxtSubst s c name e) = MetaCtxtSubst s c name (substChainVar env n e) 
 substChainVar env n (ConstantCtxtSubst s c name e) = ConstantCtxtSubst s c name (substChainVar env n e) 
 substChainVar env n (Letrec env' inexpr) = Letrec (substChainVar env n env') (substChainVar env n inexpr)    
 substChainVar env n e = e 
 
-- substEnvConstraint env n (Lam v e) = Lam v (substEnvConstraint env n e)
 substEnvConstraint env n (Fn name args) = Fn name (map (substEnvConstraint env n) args)
 substEnvConstraint env n (MetaCtxtSubst s c name e) = MetaCtxtSubst s c name (substEnvConstraint env n e) 
 substEnvConstraint env n (ConstantCtxtSubst s c name e) = ConstantCtxtSubst s c name (substEnvConstraint env n e) 
 substEnvConstraint env n (Letrec env' inexpr) = Letrec (substEnvConstraint env n env') (substEnvConstraint env n inexpr)    
 substEnvConstraint env n e = e 

instance Substitute Binding where
 subst e1 e2 (x :=  e) = (subst e1 e2 x) := (subst e1 e2 e) 
 substEmpty n (x := e) = (substEmpty n x) := (substEmpty n e) 
 substEnv env n (x := e) = (substEnv env n x) := (substEnv env n e)
 substEnvConstraint env n (x := e) = (substEnvConstraint env n x) := (substEnvConstraint env n e)
 substChainVar env n (x := e) = (substChainVar env n x) := (substChainVar env n e)
 
instance Substitute Environment where
 subst e1 e2 env = 
                    env {
                          bindings = [subst e1 e2 b | b <- bindings env]
                         ,metaChains  = [MetaChain s typ name w (subst e1 e2 e) | (MetaChain s typ name v e) <- metaChains env, let w = subst e1 e2 v ]
                         ,constantChains  = [ConstantChain s typ name w (subst e1 e2 e) | (ConstantChain s typ name v e) <- constantChains env, let w = subst e1 e2 v ]
                         }
   
 substEmpty n env = env {
                          bindings = [substEmpty n b | b <- bindings env]
                         ,metaEnvVars =     [Env s v | (Env s v) <- (metaEnvVars env), v /= n]
                                        
                         ,metaChains  = [MetaChain s typ name v (substEmpty n e) | (MetaChain s typ name v e) <- metaChains env ]
                         ,constantChains  = [ConstantChain s typ name v (substEmpty n e) | (ConstantChain s typ name v e) <- constantChains env ]
                         }
                         
 substEnv env1 (n) env 
   | (n) `elem` ([n | (Env _ n) <- metaEnvVars env]) =
                       env {
                          bindings = (bindings env1) ++ [substEnv env1 n b | b <- bindings env]
                         ,metaEnvVars =  (metaEnvVars env1) ++ [Env s v | Env s v <- (metaEnvVars env), v /= n] 
                         ,constantEnvVars = (constantEnvVars env1) ++ (constantEnvVars env)
                         ,metaChains  = (metaChains env1) ++ [MetaChain s typ name v (substEnv env1 n e) | (MetaChain s typ name v e) <- metaChains env]
                         ,constantChains  = (constantChains env1) ++ [ConstantChain s typ name v (substEnv env1 n e) | (ConstantChain s typ name v e) <- constantChains env]
                         }
   | otherwise =   env {
                          bindings = [substEnv env1 n b | b <- bindings env]
                         ,metaChains  = [MetaChain s typ name v (substEnv env1 n e) | (MetaChain s typ name v e) <- metaChains env]
                         ,constantChains  = [ConstantChain s typ name v (substEnv env1 n e) | (ConstantChain s typ name v e) <- constantChains env]
                         }
                         
 substChainVar env1 v@(MetaChain _ typ name z s) env =
    let newEnvironments = [env1 z' s' | (MetaChain s typ' name' z' s',rest) <- splits (metaChains env), name == name']
        newBindings     = concatMap bindings newEnvironments
        newChains       = concatMap metaChains newEnvironments
        newEnvVars      = concatMap metaEnvVars newEnvironments
        newConstantEnvVars = concatMap constantEnvVars newEnvironments
        newConstantChains = concatMap constantChains newEnvironments
        allBindings     = newBindings ++ [substChainVar env1 v b | b <- bindings env]
        allChains       = newChains   ++ [MetaChain s typ' name' z' (substChainVar env1 v s') | (MetaChain s typ' name' z' s') <- metaChains env, name /= name' ]
        allEnvVars      = newEnvVars ++ (metaEnvVars env)
        allConstantEnvVars = newConstantEnvVars ++ (constantEnvVars env)
        allConstantChains = newConstantChains ++ [ConstantChain s typ' name' z' (substChainVar env1 v s') | (ConstantChain s typ' name' z' s') <-constantChains env]
    in env{bindings = allBindings
          ,metaChains = allChains
          ,metaEnvVars = allEnvVars
          ,constantEnvVars = allConstantEnvVars
          ,constantChains = allConstantChains}    
                         
 substEnvConstraint env1 env2 env 
   | not $ null [n | MetaChain s t n v e <- constantChains env1] = error "metachain in constant chains env1"
   | not $ null [n | MetaChain s t n v e <- constantChains env2] = error "metachain in constant chains env2"
   | not $ null [n | MetaChain s t n v e <- constantChains env] = error "metachain in constant chains env"      
   | not $ null [n | ConstantChain s t n v e <- metaChains env1] = error "constantchain in meta chains env1"
   | not $ null [n | ConstantChain s t n v e <- metaChains env2] = error "constantchain in meta chains env2"
   | not $ null [n | ConstantChain s t n v e <- metaChains env] = error "constantchain in meta chains env"      
   | null [b | b <- bindings env2, b `notElem` (bindings env)]
   , null [b | b <- metaChains env2, b `notElem` (metaChains env)]  
   , null [b | b <- metaEnvVars env2, b `notElem` (metaEnvVars env)]  
   , null [b | b <- constantEnvVars env2, b `notElem` (constantEnvVars env)]  
   , null [b | b <- constantChains env2, b `notElem` (constantChains env)]  
     =  let res = env {
              bindings        = (bindings env1) ++ [substEnvConstraint env1 env2 b | b <- bindings env, b `notElem` (bindings env2)]
             ,metaEnvVars     = (metaEnvVars env1) ++ [v | v <- (metaEnvVars env), v `notElem` (metaEnvVars env2)]
             ,constantEnvVars = (constantEnvVars env1) ++ [c | c <- (constantEnvVars env), c `notElem` (constantEnvVars env2)]
             ,metaChains      = (metaChains env1) ++     [MetaChain s typ name v (substEnvConstraint env1 env2 e)     | c@(MetaChain s typ name v e) <- metaChains env, c `notElem` (metaChains env2)]
             ,constantChains  = (constantChains env1) ++ [ConstantChain s typ name v (substEnvConstraint env1 env2 e) | c@(ConstantChain s typ name v e) <- constantChains env, c `notElem` (constantChains env2)]
                         }
        in res
                         
                         
   | otherwise =   env {
                          bindings        = [substEnvConstraint env1 env2 b | b <- bindings env]
                         ,metaChains      = [MetaChain s typ name v (substEnvConstraint env1 env2 e)     | (MetaChain s typ name v e) <- metaChains env]
                         ,constantChains  =   [ConstantChain s typ name v (substEnvConstraint env1 env2 e) | (ConstantChain s typ name v e) <- constantChains env]
                         }
                            
-- ============================================================================


-- ============================================================================
-- Treating variable names by a type class, since there are some 'special' names 
class GetName a where
 getName :: a -> String
 hasName :: a -> Bool

instance GetName VarLift where
 getName (MetaVarLift s "--") = error "in getName, -- is not a name"
 getName (MetaVarLift s v) = v
 getName (VarLift s "--") = error "in getName, -- is not a name"
 getName (VarLift s v) = v
 getName (ConstantVarLift s v) = v
 hasName (MetaVarLift s x) = x /= "--"
 hasName (VarLift s x) = x /= "--"
 hasName (ConstantVarLift s v) = True
 

-- ============================================================================
-- Treating the meta-variables
-- ============================================================================

-- container for the different types of meta-variables

data MetaVars = MetaVars {liftvars  :: [Variable], 
                          envvars  :: [Variable],
                          ctxtvars :: [Variable],
                          exprvars :: [Variable],
                          chainvars :: [Variable]
                          }
  deriving(Show)                        

-- empty container  
emptyMVars :: MetaVars
emptyMVars = MetaVars [] [] [] [] []

-- union two containers:
combineMetaVars :: MetaVars -> MetaVars -> MetaVars
combineMetaVars m m' = 
 m {liftvars = liftvars m ++ liftvars m'
       ,envvars = envvars m ++ envvars m'
       ,chainvars = chainvars m ++ chainvars m'
       ,ctxtvars = ctxtvars m ++ ctxtvars m'
       ,exprvars = exprvars m ++ exprvars m'
       }
-- union a list of containers       
combineMetaVarsList :: [MetaVars] -> MetaVars
combineMetaVarsList list = foldr (combineMetaVars) emptyMVars list       

-- compute the meta-vars and add them to a container
addMetaVars :: HasMetaVars a => a -> MetaVars -> MetaVars
addMetaVars a m =
  let m' = computeMetaVars a
  in combineMetaVars m m'
  
-- compute the meta-vars of a list of objects  
computeMetaVarsList :: HasMetaVars a => [a] -> MetaVars
computeMetaVarsList xs = foldr (addMetaVars) emptyMVars  xs
       
-- type class for object which have meta-vars       
class HasMetaVars a where
 computeMetaVars :: a -> MetaVars
 
-- instances
instance HasMetaVars VarLift where 
 computeMetaVars (MetaVarLift _ x) = emptyMVars {liftvars = [x]}
 computeMetaVars _ = emptyMVars
 
instance HasMetaVars ArgPos where
 computeMetaVars (Binder v e)  = addMetaVars e (computeMetaVarsList v) 
 computeMetaVars (VarPos v )   = (computeMetaVars v)  

instance HasMetaVars Expression where
 computeMetaVars (VarL v)  = computeMetaVars v
 computeMetaVars (Fn f arg) =  computeMetaVarsList arg
 computeMetaVars (Letrec env e) = addMetaVars e (computeMetaVars env) 
 computeMetaVars (MetaExpression _ e) = emptyMVars {exprvars =[e]}
 computeMetaVars (Hole _) = emptyMVars
 computeMetaVars (MetaCtxt _ _  c) = emptyMVars {ctxtvars = [c]}
 computeMetaVars (MetaCtxtSubst _ _ c s) = addMetaVars s (emptyMVars {ctxtvars = [c]}) 
 computeMetaVars (Constant _ _) =emptyMVars
 computeMetaVars (ConstantCtxt _ _ _) = emptyMVars
 computeMetaVars (ConstantCtxtSubst _ _ _ s) = computeMetaVars s
 
instance HasMetaVars Binding where
 computeMetaVars (x := e) = addMetaVars e (computeMetaVars x) 
instance HasMetaVars ChainVar where
 computeMetaVars (MetaChain _ _ ee x s) = addMetaVars s (addMetaVars x (emptyMVars {chainvars = [ee]}))
 computeMetaVars (ConstantChain _ _ ee x s) = addMetaVars s (computeMetaVars x)
instance HasMetaVars Environment where
 computeMetaVars env = 
  combineMetaVarsList [computeMetaVarsList (bindings env)
                      ,emptyMVars{envvars = [v | (Env _ v) <- metaEnvVars env]}
                      ,computeMetaVarsList (metaChains env)
                      ,computeMetaVarsList (constantChains env)]                      
instance HasMetaVars a =>  HasMetaVars [a] where
    computeMetaVars l = combineMetaVarsList (map computeMetaVars l)
                      -- ============================================================================



-- ============================================================================
-- check the 'weak distinct variable convention' of an expression
  
checkScope :: Expression -> Maybe String
checkScope expr =   
  go expr [] []
 where
  goArgPos (VarPos v) scopeLam scopeLet = Nothing
  goArgPos (Binder v e)  scopeLam scopeLet = go e (v++scopeLam) scopeLet
  -- go (Lam v e)   scopeLam scopeLet = go e (v:scopeLam) scopeLet
  go (VarL _)  scopeLam scopeLet = Nothing
  go (Fn f args) scopeLam scopeLet = 
            let rec = map (\x -> goArgPos x scopeLam scopeLet) args
            in if all (isNothing) rec then Nothing else (head (filter  isJust rec))
  go (Letrec env e) scopeLam scopeLet     =
    let capvars =      [v | (MetaChain _ _ _ v _) <- metaChains env, not (isPseudoMV v)]
                    ++  [ v | (ConstantChain _ _ _ v _) <- constantChains env]
                    ++  [ b | (b := e) <- bindings env]
        expressions = e:((map (\(MetaChain  _ _ _ _ s) -> s)  $ metaChains env) 
                          ++  (map (\(ConstantChain  _ _ _ _ s) -> s)  $ constantChains env) 
                          ++  (map (\(b := e) -> e) $ bindings env))
        scopeLet' = scopeLet ++ capvars
        checkDoubleLetVar = (nub capvars) == capvars
        check     = [c | c <- capvars, c `elem` scopeLet]
        rec       = filter isJust (  map (\x -> go x scopeLam scopeLet') expressions)
    in if checkDoubleLetVar then 
          if null check then 
            if null rec then Nothing else  head (rec)
          else Nothing -- Just $ "Let-binder for variable " ++ (show $ head check) ++ "occurs in a scope of another let-binder for the same variable"
        else Just $ "Multiple let binders in the same environmet"                                                                                                                
  go (MetaExpression _ s) scopeLam scopeLet = Nothing
  go (Constant _ s) scopeLam scopeLet = Nothing
  go (Hole _) scopeLam scopeLet = Nothing
  go (MetaCtxt _ _ _) scopeLam scopeLet = Nothing
  go (ConstantCtxt _ _ _) scopeLam scopeLet = Nothing
  go (MetaCtxtSubst _ srt name expr) scopeLam scopeLet = go expr scopeLam scopeLet
  go (ConstantCtxtSubst _ srt name expr) scopeLam scopeLet = go expr scopeLam scopeLet

-- checkWeakDVC expr = Nothing
checkWeakDVC expr =
 let 
   bv = boundVars expr
   fv = freeVars expr
   liftvars =  [getName v |  v <- bv, hasName v]
   
 in case (checkScope expr) of
      Just e -> Just e
      Nothing -> 
       -- if not (allDifferent letvars) then Just $ "bound letvars aren't different: " ++ show letvars
--       else 
         let l = [v | v <- liftvars, v `elem` (map (\x -> getName x) fv)] -- bound letvars /=  freevars
         in if not (null l) then Just $ "bound vars /= freevars" ++ show l
                               else Nothing                            
-- checkWeakDVCEnv expr = Nothing
checkWeakDVCEnv env =
 let 
   bv = boundVarsEnv env
   fv = freeVarsEnv env
   liftvars =  [getName v | v <- bv, hasName v]
   expressions = ((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) 
                          ++  (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env) 
                          ++  (map (\(b := e) -> e) $ bindings env))
   dvcex = (filter isJust (map (\x -> checkWeakDVC x) expressions))                           
  in -- if not (allDifferent letvars) then Just $ "bound letvars aren't different: " ++ show letvars
      -- else  
   let l = [v | v <- liftvars, v `elem` (map (\x -> getName x) fv)] -- bound letvars /=  freevars
                    in if not (null l) then Just $ "bound vars /= freevars" ++ show l
                                       else if null dvcex then Nothing
                                                          else head dvcex
-- ============================================================================
     
-- weaker Variant for mathcing
{-
checkWeakDVCMatching expr =
 let 
   bv = boundVars expr
   fv = freeVars expr
   liftvars =  [getName v | v <- bv, hasName v]
   --if not (allDifferent letvars) then Just $ "bound letvars aren't different: " ++ show letvars
      -- else
  in case (checkScope expr) of
      Just e -> Just e
      Nothing ->       
                    let l = [v | v <- liftvars, v `elem` (map (\x -> getName x) fv)] -- bound letvars /=  freevars
                    in if not (null l) then Just $ "bound letvars /= freevars" ++ show l
                                       else Nothing                            
checkWeakDVCEnvMatching env =
 let 
   bv = boundVarsEnv env
   fv = freeVarsEnv env
   liftvars =  [getName v | v <- bv, hasName v]
   expressions = ((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) 
                          ++  (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env) 
                          ++  (map (\(b := e) -> e) $ bindings env))
   dvcex = (filter isJust (map (\x -> checkWeakDVCMatching x) expressions))      
  in --if not (allDifferent letvars) then Just $ "bound letvars aren't different: " ++ show letvars
     -- else  
                            let l = [v | v <- liftvars, v `elem` (map (\x -> getName x) fv)] -- bound liftvars /=  freevars
                            in if not (null l) then Just $ "bound liftvars /= freevars" ++ show l
                                else if null dvcex then Nothing
                                 else head dvcex            
-}     
-- ============================================================================
-- check the subTerm-Relation
-- ============================================================================
subTermOf :: Expression -> Expression -> Bool
e1 `subTermOf` e2
 | e1 == e2 = True
-- e1 `subTermOf` (Lam v e2) = e1 `subTermOf` e2
e1 `subTermOf` (Fn _ args) = any (\x -> e1 `subTermOfArgPos` x) args
e1 `subTermOf` (MetaCtxtSubst _ _ _ e2) = e1 `subTermOf` e2
e1 `subTermOf` (ConstantCtxtSubst _ _ _ e2) = e1 `subTermOf` e2
e1 `subTermOf` _ = False
e1 `subTermOfArgPos` ((Binder v e2)) = e1 `subTermOf` e2
(VarL v) `subTermOfArgPos` (VarPos v') = v == v'
e1 `subTermOfArgPos` (VarPos _) = False


-- only syntactic, i.e.  S =?= C[S] yields true, even C may represent the emtpy context
properSubTermOf :: Expression -> Expression-> Bool
properSubTermOf e1 e2  = e1 /= e2 && e1 `subTermOf` e2 

  
-- ============================================================================


-- ============================================================================
-- free and bound variables, captured variables in contexts
-- ============================================================================

-- compute all variables occuring in an expression
allVariables e = freeVars e ++ boundVars e
allVariablesInEnv env = freeVarsEnv env ++ boundVarsEnv env

-- compute the free variables, where only Meta-Variables are considered
freeVarsMetaVars :: Expression -> [VarLift]
freeVarsMetaVars e = go $ freeVars e 
  where
    go [] = []
    go (x:xs)
     | ((MetaVarLift _ _)) <- x = x:(go xs)
     | otherwise = go xs
     
-- compute the free variables, where only concrete Variables are considered
freeVarsConcrete :: Expression -> [VarLift]
freeVarsConcrete e = go $ freeVars e
  where
    go [] = []
    go (x:xs)
     | ((VarLift  _ _)) <- x = x:(go xs)
     | otherwise = go xs
     
     
-- compute the free variables, where only Constant Variables are considered
freeVarsConstant :: Expression -> [VarLift]
freeVarsConstant e = go $ freeVars e
  where
    go [] = []
    go (x:xs)
     | ((ConstantVarLift  _ _)) <- x = x:(go xs)
     | otherwise = go xs
     
      
freeVarsArgPos (VarPos v)   = [v]
freeVarsArgPos (Binder v e) = [x | x <- (freeVars e), x `notElem` (v)] 
-- compute all free variables (meta and concrete variables)      
freeVars :: Expression -> [VarLift]
-- freeVars (Lam v e) = filter (/= (Left v)) $ freeVars e
freeVars (VarL (MetaVarLift _ "--")) = [] -- hack
freeVars (VarL v) = [v]
freeVars (Fn f args) = concatMap freeVarsArgPos args
freeVars (Letrec env e) =
 let capvars =  [v | (MetaChain  _ _ _ v _) <- metaChains env, not (isPseudoMV v)]
                ++[v | (ConstantChain _ _ _ v _) <- constantChains env]
                ++ [b | (b := e) <- bindings env]
     expressions = e:((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) 
                          ++  (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env) 
                          ++  (map (\(b := e) -> e) $ bindings env))
  in [v | v <- concatMap freeVars expressions, v `notElem` capvars]
freeVars (MetaExpression _ s) = []
freeVars (Constant _ s) = []
freeVars (Hole _) = []
freeVars (MetaCtxt _  _ _) = []
freeVars (ConstantCtxt _ _ _) = []
freeVars (MetaCtxtSubst _ srt name expr) = freeVars expr     
freeVars (ConstantCtxtSubst _ srt name expr) = freeVars expr     
freeVarsEnv :: Environment -> [VarLift]
freeVarsEnv env =
 let capvars =  [v | (MetaChain _ _ _ v _) <- metaChains env, not (isPseudoMV v)]
                ++ [b | (b := e) <- bindings env]
                ++ [v | (ConstantChain _ _ _ v _) <- constantChains env]
                 
     expressions = ((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) 
                   ++ (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env) 
                   ++  (map (\(b := e) -> e) $ bindings env))
  in [v | v <- concatMap freeVars expressions, v `notElem` capvars]

-- compute all bound variables (meta and concrete variables)  
boundVarsArgPos (VarPos v) = []
boundVarsArgPos (Binder v e) = (v) ++ (boundVars e)

boundVars :: Expression -> [VarLift]
-- boundVars (Lam v e) = (Left v):boundVars e
boundVars (VarL v) = []
boundVars (Fn f args) = concatMap boundVarsArgPos args
boundVars (Letrec env e) =
 let capvars = (map (\(MetaChain _ _ _ v _) -> v)  $ metaChains env) ++  (map (\(b := e) -> b) $ bindings env)
                           ++ (map (\(ConstantChain _ _ _ v _) -> v)  $ constantChains env)
     expressions = e:((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) ++  (map (\(b := e) -> e) $ bindings env)
                     ++ (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env))
  in capvars ++ (concatMap boundVars expressions)
boundVars (MetaExpression _ s) = []
boundVars (Constant _ s) = []
boundVars (Hole _) = []
boundVars (MetaCtxt _ _ _) = []
boundVars (MetaCtxtSubst _ srt name expr) = boundVars expr     
boundVars (ConstantCtxt _ _ _) = []
boundVars (ConstantCtxtSubst _ srt name expr) = boundVars expr     
boundVarsEnv :: Environment -> [VarLift]
boundVarsEnv env =
 let capvars = (map (\(MetaChain _ _ _ v _) -> v)  $ metaChains env) ++  (map (\(b := e) -> b) $ bindings env)
                           ++ (map (\(ConstantChain _ _ _ v _) -> v)  $ constantChains env)
     expressions = ((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) ++  (map (\(b := e) -> e) $ bindings env))
                    ++ (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env)
  in capvars ++ (concatMap boundVars expressions)
  
isExpr e 
    | not (hasHole e)  = True
    | otherwise        = False
    
-- given a context, compute the variables which are captured by the hole 
capturedInHoleArgPos (Binder v e) = (v) ++ capturedInHole e
capturedInHoleArgPos (VarPos _) = []

capturedInHole :: Context -> [VarLift]
-- capturedInHole (Lam v e) = [Left v] ++ capturedInHole e
capturedInHole (VarL _) = []
capturedInHole (Fn f args) = concatMap snd $ filter (\(a,b) -> a == True) $ map (\a -> (hasHoleArgPos a, capturedInHoleArgPos a)) args
capturedInHole (Letrec env e) =
 let capvars =   (map (\(MetaChain _ _ _ v _) -> v)  $ metaChains env) ++  (map (\(b := e) -> b) $ bindings env) 
                           ++ (map (\(ConstantChain _ _ _ v _) -> v)  $ constantChains env)
     expressions = e:((map (\(MetaChain _ _ _ _ s) -> s)  $ metaChains env) ++  (map (\(b := e) -> e) $ bindings env)
                     ++ (map (\(ConstantChain _ _ _ _ s) -> s)  $ constantChains env))
  in capvars ++ (concatMap snd $ filter (\(a,b) -> a == True) $ map (\a -> (hasHole a, capturedInHole a)) expressions)
capturedInHole (MetaExpression _ s) = []
capturedInHole (Constant _ s) = []
capturedInHole (Hole _) = []
capturedInHole (MetaCtxt _ _ _) = []
capturedInHole (ConstantCtxt _ _ _) = []
capturedInHole (MetaCtxtSubst _ srt name expr) =  capturedInHole expr     
capturedInHole (ConstantCtxtSubst _ srt name expr) =  capturedInHole expr     
 
-- check whether a context has a hole (needed, since contexts are expressions)
hasHoleArgPos ((Binder v e)) = hasHole e
hasHoleArgPos ((VarPos v )) = False

hasHole :: Context -> Bool
-- hasHole (Lam v e) = hasHole e
hasHole (VarL _) = False
hasHole (Fn f args) = any hasHoleArgPos args
hasHole (Letrec env e) = hasHole e || hasHoleEnv env
hasHole (MetaExpression _ s) = False
hasHole (Constant _  _) = False
hasHole (Hole _)= True
hasHole (MetaCtxt _ _ _) = True
hasHole (ConstantCtxt _ _ _) = True
hasHole (MetaCtxtSubst _ srt name expr) = hasHole expr
hasHole (ConstantCtxtSubst _ srt name expr) = hasHole expr
hasHoleEnv env =
      (any (\(MetaChain _ _ _ _ e) -> hasHole e)  $ metaChains env)
  ||  (any (\(ConstantChain _ _ _ _ e) -> hasHole e)  $ constantChains env)
  ||  (any (\(b := e) -> hasHole e) $ bindings env)
-- ============================================================================
  
 
 

-- ============================================================================
-- class for converting in and from expressions
-- ============================================================================

class ConvExpression b where
 toExpression :: b -> Expression
 fromExpression :: Expression -> b

-- ============================================================================ 


 
-- ============================================================================ 
-- class for applying substitutions  
-- ============================================================================
class Substitute a where 
-- subst e1 e2 e3 = e3[e2/e1]
-- everything which can be converted into an expression can be used for subst 
 subst :: ConvExpression b => b -> b -> a -> a

-- substEmpty replace the environment variable named by the String by an empty
-- environment: 
 substEmpty :: String -> a -> a 
 
-- substEnv: substitutions for environments, i.e.
-- substEnv env E e = e[env/E] 
 substEnv   :: Environment -> String -> a -> a

-- substChainVar EE[z,s] env e = e[env/EE[z,s]] 
 substChainVar :: (VarLift -> Expression -> Environment) -> ChainVar -> a -> a
 substEnvConstraint :: Environment -> Environment -> a -> a
-- some instances for usual types
instance Substitute a => Substitute [a] where
  subst e1 e2 = map (subst e1 e2) 
  substEmpty n = map (substEmpty n)
  substEnv env n = map (substEnv env n)
  substChainVar v1 v2 = map (substChainVar v1 v2)  
  substEnvConstraint env n = map (substEnvConstraint env n)
            
instance (Substitute a, Substitute b) => Substitute (Either a b) where
 subst e1 e2 (Left e) = Left (subst e1 e2 e)
 subst e1 e2 (Right e) = Right (subst e1 e2 e)
 substEmpty n (Left e) = Left (substEmpty n e)
 substEmpty n (Right e) = Right (substEmpty n e)
 substEnv env n (Left e) = Left (substEnv env n e)
 substEnv env n (Right e) = Right (substEnv env n e)
 substChainVar v1 v2 (Left e) = Left (substChainVar v1 v2 e)  
 substChainVar v1 v2 (Right e) = Right (substChainVar v1 v2 e)  
 substEnvConstraint env n (Left e) = Left (substEnvConstraint env n e)
 substEnvConstraint env n (Right e) = Right (substEnvConstraint env n e)
 
instance (Substitute a, Substitute b) => Substitute (a,b) where
 subst e1 e2 (a,b) = (subst e1 e2 a,subst e1 e2 b)
 substEmpty n (a,b) = (substEmpty n a,substEmpty n b)
 substEnv e1 e2 (a,b) = (substEnv e1 e2 a,substEnv e1 e2 b)
 substChainVar v1 v2 (a,b) = (substChainVar v1 v2 a,substChainVar v1 v2 b)  
 substEnvConstraint e1 e2 (a,b) = (substEnvConstraint e1 e2 a,substEnvConstraint e1 e2 b)

-- ============================================================================ 

-- for a concrete term (or with constant metavars) t compute all 
--  t1,t2 s.t. t1[t2] = t  where t1 is a context.

splitInContexts spos ACtxt = splitInACtxt spos
splitInContexts spos TCtxt = splitInTCtxt
splitInContexts spos Ctxt = splitInCtxt
 
splitInCtxtArgPos (VarPos _) = []
splitInCtxtArgPos (Binder v e) =
  [(Binder v ctxt,expr) | (ctxt,expr) <- splitInCtxt e]
-- splitInCtxt (Lam v e) = 
  -- (Hole,Lam v e):[(Lam v ctxt,expr)   | (ctxt,expr) <- splitInCtxt e] 
splitInCtxt :: Expression -> [(Context,Expression)]
splitInCtxt (Fn f args) = 
  (Hole emptySubstitution,Fn f args):[(Fn f (before ++ [ctxt] ++ after),e')   | (before,e,after) <- splits2 args, (ctxt,e') <- splitInCtxtArgPos e]
splitInCtxt (VarL v) = [(Hole emptySubstitution,VarL v)]
splitInCtxt (MetaExpression sub s) = [(Hole emptySubstitution,MetaExpression sub s)]
splitInCtxt (Hole s) = [(Hole s,Hole s)]
splitInCtxt (MetaCtxt s sort name) =[(Hole emptySubstitution,MetaCtxt s sort name), (MetaCtxt s sort name,Hole emptySubstitution)]
splitInCtxt (ConstantCtxt s sort name) =[(Hole emptySubstitution,ConstantCtxt s sort name), (ConstantCtxt s sort name,Hole emptySubstitution)]
splitInCtxt (MetaCtxtSubst s sort name e) = 
  (Hole emptySubstitution,MetaCtxtSubst s sort name e):[(subst ctxt (Hole emptySubstitution) (MetaCtxt s sort name),expr)   | (ctxt,expr) <- splitInCtxt e] 
splitInCtxt (ConstantCtxtSubst s sort name e) = 
  (Hole emptySubstitution,ConstantCtxtSubst s sort name e):[(subst ctxt (Hole emptySubstitution) (ConstantCtxt s sort name),expr)   | (ctxt,expr) <- splitInCtxt e] 
splitInCtxt (Constant sub s) = [(Hole emptySubstitution,Constant sub s)]
splitInCtxt (Letrec env e) = 
 [(Hole emptySubstitution,Letrec env e)]
 ++
 -- hole in in-expr:
 [(Letrec env ctxt,expr)   | (ctxt,expr) <- splitInCtxt e]
 ++
 -- hole in a binding:
 [(Letrec env{bindings=(before ++ [x:=ctxt] ++ after)} e,expr) | (before, x:=e1, after) <- splits2 (bindings env), (ctxt,expr) <- splitInCtxt e1]
 ++
 -- hole in chain
 [(Letrec env{metaChains=(before ++ [MetaChain sub typ ee z ctxt] ++ after)} e,s) | (before, MetaChain sub typ ee z s, after) <- splits2 (metaChains env)
                                                                 ,(ctxt,expr) <- splitInCtxt s]
 ++
 -- hole in constant chain
 [(Letrec env{metaChains=(before ++ [ConstantChain sub typ ee z ctxt] ++ after)} e,s) | (before, ConstantChain sub typ ee z s, after) <- splits2 (metaChains env)
                                                                  ,(ctxt,expr) <- splitInCtxt s]
splitInTCtxtArgPos (Binder [] e) =
 [(Binder [] ctxt,e') | (ctxt,e') <- splitInTCtxt e]
splitInTCtxtArgPos (Binder (x:xs) e) = []
splitInTCtxtArgPos (VarPos _) = []
 
-- splitInTCtxt (Lam v e) = 
  -- [(Hole,Lam v e)]
splitInTCtxt :: Expression -> [(Context,Expression)]
splitInTCtxt (Fn f args) = 
  (Hole emptySubstitution,Fn f args):[(Fn f (before ++ [ctxt] ++ after),e')   | (before,e,after) <- splits2 args, (ctxt,e') <- splitInTCtxtArgPos e]
splitInTCtxt (VarL v) = [(Hole emptySubstitution,VarL v)]
splitInTCtxt (MetaExpression subst s) = [(Hole emptySubstitution,MetaExpression subst s)]
splitInTCtxt (Hole s) = [(Hole s,Hole s)]
splitInTCtxt (MetaCtxt sub sort name) =(Hole emptySubstitution,MetaCtxt sub sort name): if sort /= Ctxt then [(MetaCtxt sub sort name,Hole emptySubstitution)] else []
splitInTCtxt (ConstantCtxt sub sort name) =(Hole emptySubstitution,ConstantCtxt sub sort name): if sort /= Ctxt then [(ConstantCtxt sub sort name,Hole emptySubstitution)] else []
splitInTCtxt (MetaCtxtSubst sub sort name e) = 
  (Hole emptySubstitution,MetaCtxtSubst sub sort name e)
    :(if sort /= Ctxt then [(subst ctxt (Hole emptySubstitution) (MetaCtxt sub sort name),expr)   | (ctxt,expr) <- splitInTCtxt e] else [])
splitInTCtxt (ConstantCtxtSubst sub sort name e) = 
  (Hole emptySubstitution,ConstantCtxtSubst sub sort name e)
    :(if sort /= Ctxt then [(subst ctxt (Hole emptySubstitution) (ConstantCtxt sub sort name),expr)   | (ctxt,expr) <- splitInTCtxt e] else []) 
splitInTCtxt (Constant sub s) = [(Hole emptySubstitution,Constant sub s)]
splitInTCtxt (Letrec env e) = 
 [(Hole emptySubstitution,Letrec env e)]
 ++
 -- hole in in-expr:
 [(Letrec env ctxt,expr)   | (ctxt,expr) <- splitInTCtxt e]
 ++
 -- hole in a binding:
 [(Letrec env{bindings=(before ++ [x:=ctxt] ++ after)} e,expr) | (before, x:=e1, after) <- splits2 (bindings env), (ctxt,expr) <- splitInTCtxt e1]
 ++
 -- hole in chain
 [(Letrec env{metaChains=(before ++ [MetaChain sub typ ee z ctxt] ++ after)} e,s) | (before, MetaChain sub typ ee z s, after) <- splits2 (metaChains env)
                                                                 ,(ctxt,expr) <- splitInTCtxt s]
 ++
 -- hole in constant chain
 [(Letrec env{constantChains=(before ++ [ConstantChain sub typ ee z ctxt] ++ after)} e,s) 
     | (before, ConstantChain sub typ ee z s, after) <- splits2 (constantChains env)
     ,(ctxt,expr) <- splitInTCtxt s]
 
splitInACtxtArgPos spos (VarPos _) = []
splitInACtxtArgPos spos (Binder [] e) =
 [(Binder [] ctxt,e') | (ctxt,e') <- splitInACtxt spos e]
splitInACtxtArgPos _ (Binder (x:xs) e) = []
-- splitInACtxt _ (Lam v e) = [(Hole,Lam v e)]
splitInACtxt :: (String -> Maybe [Int]) -> Expression -> [(Context,Expression)]                        
splitInACtxt spos (Fn f args) 
 | Just spositions <- spos f =
   (Hole emptySubstitution,Fn f args)
                      :[(Fn f (before ++ [ctxt] ++ after),e') | 
                          (pos,(before,e,after)) <-  zip [1..] (splits2 args)
                        , (ctxt,e') <- splitInACtxtArgPos spos e
                        , pos `elem` spositions]
splitInACtxt _ (VarL v) = [(Hole emptySubstitution,VarL v)]
splitInACtxt _ (MetaExpression subst s) = [(Hole emptySubstitution,MetaExpression subst s)]
splitInACtxt _ (Hole s) = [(Hole s,Hole s)]
splitInACtxt _ (MetaCtxt s sort name) =(Hole emptySubstitution,MetaCtxt s sort name): if sort == ACtxt then [(MetaCtxt s sort name,Hole emptySubstitution)] else []
splitInACtxt _ (ConstantCtxt s sort name) =(Hole emptySubstitution,ConstantCtxt s sort name): if sort == ACtxt then [(ConstantCtxt s sort name,Hole emptySubstitution)] else []
splitInACtxt spos (MetaCtxtSubst s sort name e) = 
  (Hole emptySubstitution,MetaCtxtSubst s sort name e):(if sort == ACtxt then [(subst ctxt (Hole emptySubstitution) (MetaCtxt s sort name),expr)   | (ctxt,expr) <- splitInACtxt spos e] else [])
splitInACtxt spos (ConstantCtxtSubst s sort name e) = 
  (Hole emptySubstitution,ConstantCtxtSubst s sort name e)
  :(if sort == ACtxt 
    then [(subst ctxt (Hole emptySubstitution) (ConstantCtxt s sort name),expr)   | (ctxt,expr) <- splitInACtxt spos e] 
    else []) 
splitInACtxt _ (Constant sub s) = [(Hole emptySubstitution,Constant sub s)]
splitInACtxt _ (Letrec env e) = 
 [(Hole emptySubstitution,Letrec env e)]

nubFnSyms :: [(String,Int,[Int])] -> [(String,Int,[Int])]
nubFnSyms syms = 
 let sorted = sortBy (\(name,ar,spos) (name2,ar2,spos2) -> compare name name2) syms
     grouped = groupBy (\(name,ar,spos) (name2,ar2,spos2) -> name == name2) sorted
     handleGroup g@((n,a,s):_) = let positions = [spos | (name,ar,spos) <- g]
                     in if [] `elem` positions then (n,a,[])
                        else let l = [p | p <- positions, p /= [1], p /= []]  
                             in if null l then (n,a,s) else (n,a,head l)
 in map handleGroup grouped
              
 
class HasFnSymbol a where
 getFnSymbols :: a -> [(String,Int,[Int])]
 
instance HasFnSymbol a => HasFnSymbol [a] where
 getFnSymbols xs = concatMap getFnSymbols xs 
 
instance HasFnSymbol ArgPos where
 getFnSymbols (Binder v e) = getFnSymbols e
 getFnSymbols (VarPos v ) = []
 
instance HasFnSymbol (Expression) where
 getFnSymbols (Fn sym []) = [(sym,0,[])]
 getFnSymbols (Fn sym args) = (sym,length args,[1]):(concatMap getFnSymbols args)
 -- getFnSymbols (Lam v e)   = getFnSymbols e 
 getFnSymbols (VarL v)  = []
 getFnSymbols (Letrec env e) = (getFnSymbols env) ++ (getFnSymbols e)
 getFnSymbols (MetaExpression _ e) = []
 getFnSymbols (Hole _) = []
 getFnSymbols (MetaCtxt subst srt  c) = []
 getFnSymbols (MetaCtxtSubst subst srt c s) = getFnSymbols s
 getFnSymbols (Constant s e) = []
 getFnSymbols (ConstantCtxt s  srt  c) = [] 
 getFnSymbols (ConstantCtxtSubst _ srt c s) =  getFnSymbols s 
 
instance HasFnSymbol (Binding) where
 getFnSymbols (x := e) = (getFnSymbols e) 
instance HasFnSymbol (ChainVar) where
 getFnSymbols (MetaChain _ _ ee x s) = getFnSymbols s
 getFnSymbols (ConstantChain _ _ ee x s) = getFnSymbols s
instance HasFnSymbol (Environment) where
 getFnSymbols env = 
  (concatMap getFnSymbols (bindings env))
  ++ (concatMap getFnSymbols (metaChains env))
  ++ (concatMap getFnSymbols (constantChains env))
   
class Constantify a where 
 constantify :: a -> a
 deconstantify :: a -> a

instance Constantify a => Constantify [a] where
 constantify = map constantify
 deconstantify = map deconstantify
 
instance (Constantify a,Constantify b) => Constantify (a,b) where
 constantify (a,b) = (constantify a, constantify b)
 deconstantify (a,b) = (deconstantify a, deconstantify b)

-- instance (Constantify a,Constantify b) => Constantify (Either a b) where
--  constantify (Left a) = Left (constantify a)
--  constantify (Right a) = Right (constantify a) 
--  deconstantify (Left a) = Left (deconstantify a)
--  deconstantify (Right a) = Right (deconstantify a) 
 
 
isPseudoMV (MetaVarLift _ "--") = True
isPseudoMV _ = False

instance Constantify VarLift where
 constantify (MetaVarLift s x) = ConstantVarLift s x
 constantify e                   = e
 deconstantify (ConstantVarLift s x) = MetaVarLift s x
 deconstantify e                   = e
 
instance Constantify ArgPos where 
 constantify (VarPos v) = VarPos (constantify v)
 constantify (Binder v e) = Binder (map constantify v) (constantify e)
 deconstantify (Binder v e) = Binder (map deconstantify v) (deconstantify e)
 deconstantify (VarPos v) = VarPos (deconstantify v)
 
instance Constantify Expression where
 constantify  (VarL v) = VarL (constantify v)
 constantify  (Fn f args) = Fn f (map constantify args)
 constantify  (Letrec env e) = Letrec (constantify env) (constantify e)
 constantify  (MetaExpression subst s) = Constant subst s
 constantify  (MetaCtxt s srt name) = ConstantCtxt s srt name
 constantify  (MetaCtxtSubst s srt name (Hole _)) = ConstantCtxt s srt name 
 constantify  (MetaCtxtSubst s srt name e) = ConstantCtxtSubst s srt name (constantify e)
 constantify  (Hole s) = (Hole s)
 constantify  (Constant subst s) = Constant subst s
 constantify  (ConstantCtxt subst srt name) = ConstantCtxt subst srt name
 constantify  (ConstantCtxtSubst subst srt name (Hole _)) = ConstantCtxt subst srt name 
 constantify  (ConstantCtxtSubst subst srt name e) = ConstantCtxtSubst subst srt name (constantify e)

 deconstantify  (VarL v) = VarL (deconstantify v)
 deconstantify  (Fn f args) = Fn f (map deconstantify args)
 deconstantify  (Letrec env e) = Letrec (deconstantify env) (deconstantify e)
 deconstantify  (MetaExpression subst s) = MetaExpression subst s 
 deconstantify  (MetaCtxt s srt name) = MetaCtxt s srt name
 deconstantify  (MetaCtxtSubst s srt name e) = MetaCtxtSubst s srt name (deconstantify e)
 deconstantify  (Hole s) = (Hole s)
 deconstantify  (Constant subst s) = MetaExpression subst s
 deconstantify  (ConstantCtxt s srt name) = MetaCtxt s srt name
 deconstantify  (ConstantCtxtSubst s srt name e) = MetaCtxtSubst s srt name (deconstantify e)
 
instance Constantify Environment where
 constantify env = env {  bindings = map constantify (bindings env)
                        , metaChains = []
                        , metaEnvVars = []
                        , constantEnvVars =  (constantEnvVars env) ++ metaEnvVars env
                        , constantChains =   [ ConstantChain sub typ name (constantify z) (constantify s) | (ConstantChain sub typ name z s) <- constantChains env]
                                         ++  [ ConstantChain sub typ name (constantify z) (constantify s) | (MetaChain sub typ name z s) <- constantChains env]
                                         ++  [ ConstantChain sub typ name (constantify z) (constantify s) | (MetaChain sub typ name z s) <- metaChains env]
                                         ++  [ ConstantChain sub typ name (constantify z) (constantify s) | (ConstantChain sub typ name z s) <- metaChains env]                                         
                       }
 deconstantify env = env { bindings = map deconstantify (bindings env)
                        , constantChains = []
                        , constantEnvVars = []
                        , metaEnvVars =  (constantEnvVars env) ++ metaEnvVars env
                        , metaChains = [ MetaChain sub typ name (deconstantify z) (deconstantify s) | (ConstantChain sub typ name z s) <- constantChains env]
                                         ++  [ MetaChain sub typ name (deconstantify z) (deconstantify s) | (MetaChain sub typ name z s) <- metaChains env]
                       }
                       
instance Constantify Binding where
 constantify (b := e) = (constantify b) := (constantify e)
 deconstantify (b := e) = (deconstantify b) := (deconstantify e)  

instance (Constantify a,Constantify b) => Constantify (Either a b) where
 constantify    (Left a)        = Left  (constantify a)
 constantify    (Right a)       = Right (constantify a) 
 deconstantify  (Left a)        = Left  (deconstantify a)
 deconstantify  (Right a)       = Right (deconstantify a) 

 
type FunInfo = [(Name,Arity,[StrictPos],[BindArity])]
type Name = String
type Arity = Int
type StrictPos = Int
type BindArity = Int
