-----------------------------------------------------------------------------
-- |
-- Module      :  LRSX.Util.Util
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- several utility functions
--
-----------------------------------------------------------------------------

module LRSX.Util.Util where
import Data.Function
import qualified Data.Map.Strict as Map
import Data.List
import Data.Maybe
import Debug.Trace
import qualified Data.Set as Set

-- | 'splitBy' gets a predicate on list elements and a list
--
--  * it returns 'Nothing' if no element fulfills the predicate
--
--  * it returns 'Just' (x,xs) if x is the first element in the
--    list that fulfills the predicate and xs is the list without x
splitBy :: (a -> Bool) -> [a] -> Maybe (a,[a])
splitBy p [] = Nothing    
splitBy p (x:xs)
 | p x = Just (x,xs)
 | otherwise =
     case splitBy p xs of
       Just (y,ys) -> Just (y,(x:ys))     
       Nothing -> Nothing
       
       
-- | 'splits' splits a list into pairs (x,xs) where x is an element and xs are 
-- all other elements of the list       

splits :: [a] -> [(a,[a])]
splits list = 
 [ let (xs,x:ys) = splitAt i list in (x,(xs++ys)) | i <- [0..length list-1]]

-- | 'splits2' splits a list xs into 3-tupels (before,x,after), s.t. 
-- @xs = before ++ [x] ++ after@
--
-- e.g. @splits2 [1,2,3] = [([],1,[2,3]),([1],2,[3]),([1,2],3,[])]@
splits2 :: [a] -> [([a],a,[a])] 
splits2 list = 
 [ let (xs,x:ys) = splitAt i list in (xs,x,ys) | i <- [0..length list-1]]
 
-- | 'splits3' splits a list xs into 3-tupels (before,x,after), s.t. 
-- @xs = before ++ [x] ++ after@, but in reversed order compared to 'splits2':
--
-- e.g. @splits3 [1,2,3] = [([1,2],3,[]),([1],2,[3]),([],1,[2,3])]@
splits3 :: [a] -> [([a],a,[a])] 
splits3 list = 
 [ let (xs,x:ys) = splitAt i list in (xs,x,ys) | i <- [length list-1,length list-2..0]]
 
-- |  'assign' xs ys computes all assignments of elements of ys to elements of xs
-- the result is a list of assignments
-- each assignment is a list of pairs (x,ys') where the first components x
-- are the list xs and ys' is the list of assigned elements of ys 
-- 
-- example:
--
-- @
-- assign [1,2] ['A','B'] =
--  [[(1,['A','B']),(2,[])]
--  ,[(1,['A']),(2,['B'])]
--  ,[(1,['B']),(2,['A'])]
--  ,[(1,[]),(2,['A','B'])]]
-- @
assign :: [a] -> [b] -> [[(a,[b])]]
assign xs ys        = 
 let n = length xs
     nums = partitionInN n ys
     numberedxs = zip [1..] xs
     assignments = map (map (\x -> (fst (head x),map snd x))) $ map (groupBy (\x y -> fst x == fst y)) $ map (sortBy (compare `on` fst))  nums
  in    [merge numberedxs ((sortBy (compare `on` fst) (asgn))) |    asgn <- assignments]     
 where
    merge ((i,a):xs) yss@((j,b):ys)
     | i == j  = (a,b):(merge xs ys)
     | otherwise = (a,[]):merge xs yss
    merge ((i,a):xs) [] = (a,[]):(merge xs [])
    merge _ _ = [] 
    
partitionInN n objs = [zip xs objs | xs <- gen n (length objs)] 
gen n 0   = [[]]  
gen n len = [  i:xs | i <- [1..n], xs <- (gen n (len-1))]

-- | 'replacePos' i e xs replaces the element at position i in xs by e
-- note that the first element has position 1 (not 0!)
replacePos :: Int -> b -> [b] -> [b]
replacePos i e list = 
 let (xs,a:as) = splitAt (i-1) list
 in (xs ++ (e:as))

-- ''isSingleton' checks whether a list has exactly one element
isSingleton :: [a] -> Bool
isSingleton [x] = True
isSingleton _ = False

-- 'allDifferent' checks whether all elements of a list a different
allDifferent :: Ord a => [a] -> Bool
allDifferent xs = go (sort xs)
 where
  go [a] = True
  go []  = True
  go (a:b:xs)
   | a == b = False
   | otherwise = go (b:xs)   


   
  
findJoinPointsBy' predicate list1 singleton =
 go [] (zip list1 [1..]) 
 where 
    a = head singleton 
    go before ((l,x):list)
        | l `predicate` a = (reverse (l:before),singleton):(go (l:before) list)
        | otherwise = go (l:before) list
    go before [] = [] 
  
    
findJoinPointsBy predicate list1 list2 =
 let
  joinAtLength l list1 list2 =
   let list1p = take l list1 
       list2p = take l list2
       joins  = [(list1prefix,list2prefix)
                  | list1prefix <- inits list1p
                  , list2prefix <- inits list2p
                  , not (null list1prefix)
                  , not (null list2prefix)
                  , predicate (last list1prefix) (last list2prefix)
                  , length (list1prefix) + length (list2prefix) == l
                 ]
   in if null joins then Nothing else Just joins
  maxLength = (length list1) + (length list2)                          
 in concat  
     [fromJust joins | i <- [1..maxLength]
            ,let joins = joinAtLength i list1 list2
            ,isJust joins]
            
     
listEq xs ys = (and [x `elem` ys | x <- xs]) && (and [y `elem` xs | y <- ys])

fst4 (a,b,c,d) = a
snd4 (a,b,c,d) = b
trd4 (a,b,c,d) = c
fth4 (a,b,c,d) = d
fastLazyNubOn p xs = go p xs (Set.empty)
 where 
   go p [] _ = []
   go p (x:xs) mp 
     | Set.member (p x) mp = go p xs mp

     | otherwise = x:(go p xs (Set.insert (p x) mp))

fastLazyNub xs = go xs (Set.empty)
 where 
   go [] _ = []
   go (x:xs) set
     | x `Set.member` set = go xs set
     | otherwise = x:(go xs (Set.insert x set))
     
lookupPrefix input []   z s = Nothing
lookupPrefix input ((k,r):xs) z s
 | k `isPrefixOf` input = let rest =(drop (length k) input) in Just (r (z,s),length k,rest)
 | otherwise           = lookupPrefix input xs z s     
 
 
-- iterative deepening

dfsWithBound  :: Integral b => (a -> Bool)    -- Zieltest (goal)
     -> (a -> [a])     -- Nachfolgerfunktion (succ)
     -> [a]            -- Liste der zu bearbeitenden Knoten
     -> b        -- Tiefe
     -> Maybe (a, [a]) -- Ergebnis: Just (Zielknoten,Pfad) oder Nothing
     
dfsWithBound goal succ stack maxdepth = 
  go   [(k,maxdepth,[k]) | k <- stack] 
 where
  go [] = Nothing -- nothing found
  go ((k,i,p):r) 
   | goal k = Just (k,p) -- found goal
   | i > 0  = go ([(k',i -1, k':p) | k' <- succ k] ++ r)  
   | otherwise =  go r -- bound reached
             
idfs :: Integral b => 
        b   -- Bound
        -> (a -> Bool) -- goal test
                -> (a -> [a])  -- successor function 
                -> [a]         -- current list l
                -> Maybe (a,[a])     -- Resutlt: Just (goal,path) or Nothing
                
idfs maxbound goal succ stack = 
 let -- compute all searches until the bound exceeds
     allSearches = [dfsWithBound goal succ stack i | i <- [1..maxbound]]
 in
   case filter isJust allSearches of
     [] -> Nothing -- Nothing found
     (r:rs) -> r   -- return first result

divide [] = [([],[])]
divide (x:xs) = 
  [(x:ys,zs) | (ys,zs) <- divide xs]
  ++
  [(ys,x:zs) | (ys,zs) <- divide xs]
  
  
minimizeSets listOflists =
  let listOfSetsWD = map Set.fromList listOflists
      setOfSets = Set.fromList listOfSetsWD
      listOfSets = Set.elems setOfSets
  in [Set.elems set | (set,rest) <- splits listOfSets, all (== False) [(r `Set.isSubsetOf` set) | r <- rest]]

-- !!! is like !! but with better error message  
xs !!! i = if length xs <= i then error $ "!!! " ++ show xs ++ show i
            else xs !! i
  
lookupBy p x xs = case  [y | y <- xs, p x y] of
                    [] -> Nothing
                    (z:_) -> Just z
    
lookupOrId key mp = 
    case Map.lookup key mp of
         Nothing -> key
         Just v -> v
         
lookupAll key list = case [v | (k,v) <- list, k == key] of
                      [] -> Nothing
                      vals -> Just vals
