-----------------------------------------------------------------------------
-- |
-- Module      :  Main
-- Copyright   :  (c) D. Sabel, Goethe-University, Frankfurt a.M., Germany
-- License     :  BSD-style 
-- 
-- Maintainer  :  sabel <at> ki.cs.uni-frankfurt.de
--
--  
-- The Main-module of the LRSX Unification tool, mainly handles input and
-- output and some pretty printing.
--
-----------------------------------------------------------------------------
module Main where
import Data.List
import System.Environment
import Data.Maybe
import Data.Either
import System.Process
import Text.Read

-- import the stuff of the LRSX libraries
import LRSX.Interface.Pretty
import LRSX.Language.Syntax
import LRSX.Unification.UnificationProblem
import qualified LRSX.Unification.Unification as UN
import LRSX.Language.Constraints
import LRSX.Interface.Parser
import LRSX.Interface.DocStrings
import LRSX.Interface.Lexer
import LRSX.Util.Util
import LRSX.Diagrams.Rules
import System.IO

-- * The main function

-- | The main function
--   First get the arguments from the command line, the parse the arguments and then process the parsed options
main :: IO ()
main = do 
    -- get arguments from the command line
    args <- getArgs
    -- first parse arguments into options of type 'OptCommand', then process the options
    processOpts $ parseArgs args

-- * Handling of Options
-- Several data types and functions to parse and process the options as given by the input
    
-- | Data type for the options of the main module    
data OptCommand = Match       [SimpleOpt]  -- ^ match equations
                | Unify       [SimpleOpt]  -- ^ unify equations
                | Overlap     [SimpleOpt]  -- ^ compute critical pairs 
                | Help String              -- ^ print help text, the addtional 'String' argument is for an error message
                | OverlapAndJoin        [OptDiagram] -- ^ compute forking / commuting diagrams

 deriving(Eq,Show)
                
-- 
-- ** Parsing options                

-- | 'parseArgs' parses the command line arguments into 'OptCommand'                
parseArgs []               = Help    "No command given"
parseArgs ("join":rest)    = parseOpt (OverlapAndJoin []) rest []  -- not implemented yet
parseArgs ("unify":rest)   = parseSOpt (Unify []) rest []
parseArgs ("match":rest)   = parseSOpt (Match []) rest []    -- not implemented yet
parseArgs ("overlap":rest) = parseSOpt (Overlap []) rest []
parseArgs ("help":rest)    = Help ""
parseArgs (a:xs)           = Help ("Unknown command: " ++ a)
    
-- | parse sub-options for join mode 
parseOpt (OverlapAndJoin _) [] parsed =  error "not implemented yet"
              
parseOpt a  ("short":rest) parsed  = parseOpt a rest (parsed ++ [Short])
parseOpt a  ("long":rest) parsed   = parseOpt a rest (parsed ++ [Long]  )
parseOpt a  ("proofs":rest) parsed = parseOpt a rest (parsed ++ [Proofs] ) 
parseOpt a  ("debug":rest) parsed  = parseOpt a rest (parsed ++ [DebugOpt]  )
parseOpt a  ("help":rest) parsed   = parseOpt a rest (parsed ++ [HelpOpt ""])
parseOpt a  [infile]       parsed =  parseOpt a []   (parsed ++ [Infile infile])
parseOpt a  (other:rest) parsed = parseOpt a rest (parsed ++ [HelpOpt ("unknow flag: " ++ other)])

-- | parse sub-options for unify and overlap mode
parseSOpt (Unify _) [] parsed    = 
    let noInfile = null [() | (SInfile _) <- parsed]
    in  if noInfile 
          then Unify [SHelpOpt ("No input given")]
          else Unify parsed

parseSOpt (Match _) [] parsed    = 
    let noInfile = null [() | (SInfile _) <- parsed]
    in  if noInfile 
          then Match [SHelpOpt ("No input given")]
          else Match parsed

parseSOpt (Overlap _) [] parsed    = 
    let noInfile = null [() | (SInfile _) <- parsed]
    in   if noInfile 
            then Overlap [SHelpOpt ("No input given")]
            else Overlap parsed

parseSOpt a ("help":rest) parsed = parseSOpt a rest (parsed ++ [SHelpOpt ""])
parseSOpt a ("debug":rest) parsed = parseSOpt a rest (parsed ++ [SDebug])
parseSOpt a [infile] parsed = parseSOpt a [] (parsed ++ [SInfile infile])
parseSOpt a (other:rest) parsed = parseSOpt a rest (parsed ++ [SHelpOpt ("unknow flag: " ++ other)])
              

-- | The data type 'OptDiagram' represents sub-options for diagram computation
data OptDiagram = 
      Short            -- ^ short ASCII-representation of the diagrams, diagrams are compressed, does not work well, if diagrams aren't closed.
    | Long             -- ^ print all diagrams in ASCII-representation, no compression, works also well if diagrams cannot be closed.
    | Proofs           -- ^ print diagrams and proofs which show the critical pairs and the joining, no compression, works well if diagrams cannot be closed
    | Infile String    -- ^ the input file name
    | HelpOpt String   -- ^ help for the specific command
    | DebugOpt         -- ^ option for turning on the debug mode
 deriving(Eq,Show)
 
-- | show an 'OptDiagram'-values as 'String' 
showOptDiagram :: OptDiagram -> String
showOptDiagram Short       = "short" 
showOptDiagram Long        = "long" 
showOptDiagram Proofs      = "proofs" 
showOptDiagram DebugOpt    = "debug" 
showOptDiagram (Infile f)  = f 
showOptDiagram (HelpOpt s) = s 

-- | Type for simple options
data SimpleOpt = SInfile String  -- ^ the input file name
               | SHelpOpt String -- ^ help for the specific command
               | SDebug          -- ^ noisy output
 deriving(Eq,Show)

-- | decides whether the option is exclusive and cannot be used
--   with other options (except for input and bound)
isSingleOpt:: OptDiagram -> Bool
isSingleOpt Short  = True
isSingleOpt Long   = True
isSingleOpt Proofs = True
isSingleOpt _      = False

-- | decides whether the option is the help option and returns the 'String' within the help-option
isHelpOpt :: OptDiagram -> Maybe String
isHelpOpt (HelpOpt s) = Just s
isHelpOpt _ = Nothing
-- | decides whether the option is the help option and returns the 'String' within the help-option
isSHelpOpt :: SimpleOpt -> Maybe String
isSHelpOpt (SHelpOpt s) = Just s
isSHelpOpt _ = Nothing


-- ** Processing Options
-- | process the options, i.e. execute the corresponding parts of the LRSX Unification Tool
processOpts :: OptCommand -> IO ()
-- ---------------------------
-- computing  diagrams:
-- ---------------------------
processOpts (OverlapAndJoin opts) = error "not implemented yet"
-- ---------------------------
-- unification of expressions
-- ---------------------------
processOpts (Unify opts)
    | (Just str):_ <- filter isJust (map isSHelpOpt opts) =
        do 
         putStrLn str  
         putStrLn (commandHelp "unify")
    | otherwise                      =
        let infile = head [file | SInfile file <- opts]
            debug  = not (null [() | SDebug <- opts])
        in unifyInputEquations debug infile
        
processOpts (Match opts) = putStrLn "Not supported yet"        
-- ---------------------------
-- computing critical pairs
-- ---------------------------
processOpts (Overlap opts)
    | (Just str):_ <- filter isJust (map isSHelpOpt opts) =
        do 
         putStrLn str  
         putStrLn (commandHelp "overlap")
    | otherwise                      =
        let infile =head [file | SInfile file <- opts]
            debug  = not (null [() | SDebug <- opts])            
        in overlapAndComputeCriticalPairs debug (infile)
-- ---------------------------
-- help mode:
-- ---------------------------
processOpts (Help str) = 
    do
         putStrLn str  
         putStrLn (globalHelp)
        
-- * Several "main"-functions for the different modi        

-- ** Unifying equations with constraints
unifyInputEquations :: Bool         -- ^ flag whether debug mode is turned on (True) or not (False)
                        -> FilePath -- ^ path to input file
                        -> IO ()
unifyInputEquations debug infile = 
  do
    inp <- importPrePocessor infile
    let ((eqs,ctxtcon,envcon,capcon),finfo0) = parser inp
    let finfo = compactFunInfo finfo0
    let capconstr = (capcon) ++ concat [computeConstraints l ++ computeConstraints r     | (l:=?=:r) <-  eqs]
    let sol =[]
    let gam =(Gamma {varLiftEq = []
                    , expressionEq =  eqs
                   , environmentEq = []
                   , bindingEq = []
                   }) 
    let del = (Delta {delta1=ctxtcon,delta2=envcon,delta3=capconstr,delta4=emptyGamma}) 
    let frsh = ["_#" ++ show i | i <- [1..]]     
    let fn = nubFnSyms (getFnSymbols eqs)
    let prob = Problem {  inputGamma=gam
                        , inputDelta=del
                        , solution=sol
                        , gamma=gam
                        , delta=del
                        , fresh=frsh
                        , functions=finfo
                        , constantDelta=emptyDelta
                        , expandAllContexts=False
                        , alphaBase = []
                        , mpLogging = []
                        }
    case check prob of 
       Just err -> error err -- "Weak DVC failure in input"
       Nothing  ->
         do 
           let res = UN.solvedAndOpenLeaves $ UN.unification prob
           -- pretty print the recognized function symbols
           putStrLn (showStatisticFunInfo finfo)
           -- pretty print the input problem
           putStrLn "==============================="
           putStrLn "Input Problem:"
           putStrLn "==============================="
           putStrLn (showInputUnificationProblem prob)
           putStrLn "==============================="
           -- pretty print the unifiers
           putStrLn (showUnificationResult res)
           if debug 
              then do
                    let fulltree = UN.unification prob 
                    putStrLn "Unification Tableau (debug):"    
                    putStrLn "============================"                       
                    print fulltree
              else return ()
--         
-- ** Compute critical pairs, and display them on stdout
overlapAndComputeCriticalPairs :: Bool      -- ^ flag whether debug mode is turned on (True) or not (False)
                                -> FilePath -- ^ path to input file
                                -> IO ()
overlapAndComputeCriticalPairs debug file =
  do
   content <- importPrePocessor file
   let (((stdRules,transRules,answerRules),commands,unions),finfo0) = parseSRulesAndCommands $ lexInput content
   let finfo = compactFunInfo finfo0
   let fresh = ["#"++ show i | i <- [1..]]
   let commands'                                        = unfoldAll stdRules answerRules commands
   let overlapProblems = 
        [OverlapProblem { rules = (stdRules++[r{standardReduction=True} | r <- answerRules]++transRules)
                        , command = cmd
                        , freshNames = fresh
                        , funInfo = finfo
                        , criticalPairs = [] -- will be computed
                        , isCommuting=False
                        , opLogging=[]
                        , answerOverlap = isAnswerRuleCommand answerRules cmd
                        }
                    | cmd <- commands']
   let result = executeCommands overlapProblems
   -- one critical pair per problem
   let flattenedResult = [overlapPB{criticalPairs = [cp]}  | overlapPB <- result, cp <- criticalPairs overlapPB]
   let groupedResult = [(gp!!0){criticalPairs = concatMap criticalPairs gp}
                        | gp <- 
                            groupBy 
                                (\overlapPB1 overlapPB2 -> extractTrans (command overlapPB1) == extractTrans (command overlapPB2)) 
                                $ sortBy (\overlapPB1 overlapPB2 -> (extractTrans (command overlapPB1)) `compare` (extractTrans (command overlapPB2)))
                                    flattenedResult]
   putStrLn "========================================================="
   putStrLn $ "Computing overlaps from file " ++  file
   putStrLn "---------------------------------------------------------"
   putStrLn $ (showStatisticFunInfo finfo)
   putStrLn "---------------------------------------------------------"
   putStrLn $ unlines $ map (ppCriticalPair . head . criticalPairs) flattenedResult
   putStrLn "========================================================="
   putStrLn "Statistics: "
   putStrLn "---------------------------------------------------------"
   putStrLn $ (fill 50 " ") ++ "#critical pairs"
   putStrLn $ unlines $ map (\overlapProblem -> (fill 50 
                                                 (case command overlapProblem of
                                                    Single (Left r1) (Left r2)  _  ->  "<==" ++ showRuleName r1 ++ "== . ==" ++ showRuleName r2 ++ "==>"
                                                    Single (Right r1) (Left r2) _  ->  "==" ++ showRuleName r1 ++ "==> . ==" ++ showRuleName r2 ++ "==>"                                                 
                                                    Single (Left r1) (Right r2) _  ->  "<==" ++ showRuleName r1 ++ "== . <==" ++ showRuleName r2 ++ "=="
                                                    Single (Right r1) (Right r2) _  ->  "==" ++ showRuleName r1 ++ "==> . <==" ++ showRuleName r2 ++ "=="                                                 
                                                 )
                                                )
                                                ++ (": " ++ (show $ length $ criticalPairs overlapProblem)))
                            result
   putStrLn "---------------------------------------------------------"
   putStrLn "Overlaps per transformation"
   putStrLn $ (fill 50 " ") ++ "#critical pairs"
   putStrLn $ unlines $ map (\overlapProblem -> (fill 50 (extractTrans (command overlapProblem))) ++ (": " ++ (show $ length $ criticalPairs overlapProblem)))  groupedResult
   putStrLn "---------------------------------------------------------"
   putStrLn "========================================================="
   putStrLn $ (fill 50 " Total number of critical pairs: ") ++ (show $  length flattenedResult)
   putStrLn "========================================================="   
 where 
    extractTrans (Single (Left r)  _ _) = showRuleName r
    extractTrans (Single (Right r) _ _) = showRuleName r    




